from tests.wavefunction.test_slaterjastrow import TestSlaterJastrow
import unittest

import numpy as np
import torch
import torch.optim as optim

from qmctorch.sampler import Metropolis
from qmctorch.solver import Solver
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrowBackFlow, SlaterJastrow
from qmctorch.utils import set_torch_double_precision


def reset_generator():
    torch.manual_seed(0)
    np.random.seed(0)


class TestCompareLiHBackFlowPySCF(unittest.TestCase):

    def setUp(self):

        set_torch_double_precision()
        reset_generator()

        # molecule
        self.mol = Molecule(
            atom='Li 0 0 0; H 0 0 3.015',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        # molecule
        self.mol_ref = Molecule(
            atom='Li 0 0 0; H 0 0 3.015',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        # backflow wave function
        self.wf = SlaterJastrowBackFlow(self.mol,
                                        kinetic='jacobi',
                                        configs='single_double(2,2)',
                                        include_all_mo=True)
        self.wf.ao.backflow_trans.backflow_kernel.weight.data *= 0.
        self.wf.ao.backflow_trans.backflow_kernel.weight.requires_grad = False

        # normal wave function
        self.wf_ref = SlaterJastrow(self.mol_ref,
                                    kinetic='jacobi',
                                    include_all_mo=True,
                                    configs='single_double(2,2)')

        # fc weights
        self.random_fc_weight = torch.rand(self.wf.fc.weight.shape)
        self.wf.fc.weight.data = self.random_fc_weight.clone()
        self.wf_ref.fc.weight.data = self.random_fc_weight.clone()

        # jastrow weights
        self.random_jastrow_weight = torch.rand(
            self.wf.jastrow.jastrow_kernel.weight.shape)

        self.wf.jastrow.jastrow_kernel.weight.data = self.random_jastrow_weight.clone()
        self.wf_ref.jastrow.jastrow_kernel.weight.data = self.random_jastrow_weight.clone()

        reset_generator()
        # sampler
        self.sampler = Metropolis(
            nwalkers=5,
            nstep=200,
            step_size=0.05,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        reset_generator()
        self.sampler_ref = Metropolis(
            nwalkers=5,
            nstep=200,
            step_size=0.05,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        # optimizer
        reset_generator()
        self.opt = optim.Adam(self.wf.parameters(), lr=0.01)

        reset_generator()
        self.opt_ref = optim.Adam(self.wf_ref.parameters(), lr=0.01)

        # solver
        self.solver_ref = Solver(wf=self.wf_ref, sampler=self.sampler_ref,
                                              optimizer=self.opt_ref)

        self.solver = Solver(wf=self.wf, sampler=self.sampler,
                                          optimizer=self.opt)

        # artificial pos
        self.nbatch = 10
        self.pos = torch.as_tensor(np.random.rand(
            self.nbatch, self.wf.nelec*3))
        self.pos.requires_grad = True

    def test_0_wavefunction(self):

        # compute the kinetic energy using bf orb
        reset_generator()
        e_bf = self.wf.kinetic_energy_jacobi(self.pos)

        # compute the kinetic energy
        reset_generator()
        e_ref = self.wf_ref.kinetic_energy_jacobi(self.pos)

        print(torch.stack([e_bf, e_ref], axis=1).squeeze())
        assert torch.allclose(
            e_bf.data, e_ref.data, rtol=1E-4, atol=1E-4)

    def test1_single_point(self):

        # sample and compute observables
        reset_generator()
        obs = self.solver.single_point()
        e_bf, v_bf = obs.energy, obs.variance

        obs = self.solver.single_point()
        e_bf, v_bf = obs.energy, obs.variance

        # sample and compute observables
        reset_generator()
        obs_ref = self.solver_ref.single_point()
        e_ref, v_ref = obs_ref.energy, obs.variance

        obs_ref = self.solver_ref.single_point()
        e_ref, v_ref = obs_ref.energy, obs.variance

        # compare values
        assert torch.allclose(
            e_bf.data, e_ref.data, rtol=1E-4, atol=1E-4)

        assert torch.allclose(
            v_bf.data, v_ref.data, rtol=1E-4, atol=1E-4)

    def test2_wf_opt_grad_auto(self):

        nepoch = 5

        # optimize using backflow
        self.solver.configure(track=['local_energy'],
                              loss='energy', grad='auto')
        self.solver.configure_resampling(mode='never')

        reset_generator()
        obs = self.solver.run(nepoch)
        e_bf = torch.as_tensor(np.array(obs.energy))

        # optimize using ref
        self.solver_ref.configure(track=['local_energy'],
                                  loss='energy', grad='auto')
        self.solver_ref.configure_resampling(mode='never')

        reset_generator()
        obs_ref = self.solver_ref.run(nepoch)
        e_ref = torch.as_tensor(np.array(obs_ref.energy))

        assert torch.allclose(
            e_bf, e_ref, rtol=1E-4, atol=1E-4)

    def test3_wf_opt_grad_manual(self):

        nepoch = 5

        # optimize using backflow
        reset_generator()
        self.solver.configure(track=['local_energy', 'parameters'],
                              loss='energy', grad='manual')
        obs = self.solver.run(nepoch)
        e_bf = torch.as_tensor(np.array(obs.energy))

        # optimize using backflow
        reset_generator()
        self.solver_ref.configure(track=['local_energy', 'parameters'],
                                  loss='energy', grad='manual')
        obs = self.solver_ref.run(nepoch)
        e_ref = torch.as_tensor(np.array(obs.energy))

        # compare values
        assert torch.allclose(
            e_bf, e_ref, rtol=1E-4, atol=1E-4)


if __name__ == "__main__":
    # unittest.main()
    t = TestCompareLiHBackFlowPySCF()
    t.setUp()
    t.test_0_wavefunction()
    t.test1_single_point()
    t.test2_wf_opt_grad_auto()
    t.test3_wf_opt_grad_manual()



import unittest
import horovod.torch as hvd
import torch


class TestHVD(unittest.TestCase):
    def setUp(self):
        hvd.init()

    def test_mpi_support(self):
        assert hvd.mpi_threads_supported()

    def test_horovod_allreduce(self):
        rank = hvd.rank()
        size = hvd.size()

        tensor = torch.zeros(size)
        tensor[rank] = 1
        tensor = hvd.allreduce(tensor, op=hvd.Sum)
        assert (tensor == torch.ones(size)).all()


if __name__ == "__main__":
    t = TestHVD()
    t.setUp()
    t.test_mpi_support()
    t.test_horovod_allreduce()



from typing import Dict

import torch
from tqdm import tqdm

from .sampler_base import SamplerBase
from .. import log


class Hamiltonian(SamplerBase):

    def __init__(self,
                 nwalkers: int = 100,
                 nstep: int = 100,
                 step_size: float = 0.2,
                 L: int = 10,
                 ntherm: int = -1,
                 ndecor: int = 1,
                 nelec: int = 1,
                 ndim: int = 3,
                 init: Dict = {'min': -5, 'max': 5},
                 cuda: bool = False):
        """Hamiltonian Monte Carlo Sampler.

        Args:
            nwalkers (int, optional): Number of walkers. Defaults to 100.
            nstep (int, optional): Number of steps. Defaults to 100.
            step_size (int, optional): length of the step. Defaults to 0.2.
            L (int, optional): length of the trajectory . Defaults to 10.
            nelec (int, optional): total number of electrons. Defaults to 1.
            ntherm (int, optional): number of mc step to thermalize. Defaults to -1, i.e. keep only last position
            ndecor (int, optional): number of mc step for decorrelation. Defaults to 1.
            ndim (int, optional): total number of dimension. Defaults to 3.
            init (dict, optional): method to init the positions of the walkers. See Molecule.domain()
            cuda (bool, optional): turn CUDA ON/OFF. Defaults to False.
        """

        SamplerBase.__init__(self, nwalkers, nstep,
                             step_size, ntherm, ndecor,
                             nelec, ndim, init, cuda)
        self.traj_length = L

    @staticmethod
    def get_grad(func, inp):
        """get the gradient of the pdf using autograd

        Args:
            func (callable): function to compute the pdf
            inp (torch.tensor): input of the function

        Returns:
            torch.tensor: gradients of the wavefunction
        """
        with torch.enable_grad():
            if inp.grad is not None:
                inp.grad.zero_()
            inp.requires_grad = True

            val = func(inp)
            val.backward(torch.ones(val.shape))

            inp.requires_grad = False
        return inp.grad

    @staticmethod
    def log_func(func):
        """Compute the negative log of  a function

        Args:
            func (callable): input function

        Returns:
            callable: negative log of the function
        """
        return lambda x: -torch.log(func(x))

    def __call__(self, pdf, pos=None, with_tqdm=True):
        """Generate walkers following HMC

        Arguments:
            pdf {callable} -- density to sample
            pos (torch.tensor): precalculated position to start with
            with_tqdm (bool, optional): use tqdm progress bar. Defaults to True.

        Returns:
            torch.tensor -- sampling points
        """

        if self.ntherm < 0:
            self.ntherm = self.nstep + self.ntherm

        self.walkers.initialize(pos=pos)
        self.walkers.pos = self.walkers.pos.clone()

        # get the logpdf function
        logpdf = self.log_func(pdf)

        pos = []
        rate = 0
        idecor = 0

        rng = tqdm(range(self.nstep),
                   desc='INFO:QMCTorch|  Sampling',
                   disable=not with_tqdm)

        for istep in rng:

            # move the walkers
            self.walkers.pos, _r = self._step(
                logpdf, self.get_grad, self.step_size, self.traj_length,
                self.walkers.pos)
            rate += _r

            # store
            if istep >= self.ntherm:
                if idecor % self.ndecor == 0:
                    pos.append(self.walkers.pos)
                idecor += 1

        # print stats
        log.options(style='percent').debug(
            "  Acceptance rate %1.3f %%" % (rate / self.nstep * 100))
        return torch.cat(pos).requires_grad_()

    @staticmethod
    def _step(U, get_grad, epsilon, L, q_init):
        """Take one step of the sampler

        Args:
            U (callable): the target pdf
            get_grad (callable) : get the value of the target dist gradient
            epsilon (float) : step size
            L (int) : number of steps in the traj
            q_init (torch.Tensor) : initial positon of the walkers

        Returns:
            torch.tensor, float:
        """

        q = q_init.clone()

        # init the momentum
        p = torch.randn(q.shape)

        # initial energy terms
        E_init = U(q) + 0.5 * (p*p).sum(1)

        # half step in momentum space
        p -= 0.5 * epsilon * get_grad(U, q)

        # full steps in q and p space
        for iL in range(L - 1):
            q += epsilon * p
            p -= epsilon * get_grad(U, q)

        # last full step in pos space
        q += epsilon * p

        # half step in momentum space
        p -= 0.5 * epsilon * get_grad(U, q)

        # negate momentum
        p = -p

        # current energy term
        E_new = U(q) + 0.5 * (p*p).sum(1)

        # metropolis accept/reject
        eps = torch.rand(E_new.shape)
        rejected = (torch.exp(E_init - E_new) < eps)
        q[rejected] = q_init[rejected]

        # compute the accept rate
        rate = 1 - rejected.sum().float() / rejected.shape[0]

        return q, rate



import os
import unittest

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch.autograd import Variable

from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow

from ...path_utils import PATH_TEST

__PLOT__ = False


def read_cubefile(fname):
    with open(fname, 'r') as f:
        data = f.readlines()
    vals = []
    for d in data[7:]:
        vals.append(float(d.split('\n')[0]))
    return vals


def get_pts(npts):
    x = torch.linspace(-1, 1, npts)
    xx, yy = torch.meshgrid((x, x))
    pts = torch.cat((xx.reshape(-1, 1), yy.reshape(-1, 1)), axis=1)
    pts = torch.cat((pts, torch.zeros(pts.shape[0], 1)), axis=1)
    return pts


def generate_cube_files(t21file, npts):
    create_densf_input_file(t21file, npts)
    os.system('$ADFBIN/densf < densf_input')


def create_densf_input_file(t21name, npts):

    f = open('densf_input', 'w')
    f.write('INPUTFILE %s\n\nCUBOUTPUT MO_\n\n' % t21name)

    f.write('GRID \n')
    f.write(' -1 -1 0\n')
    f.write(' %d %d\n' % (npts, npts))
    f.write(' 1 0 0 2\n')
    f.write(' 0 1 0 2\n')
    f.write('END\n\n')

    f.write('Orbitals SCF\n')
    f.write('    A occ\n')
    f.write('    A virt\n')
    f.write('End\n\n')


class TestMOvaluesADF(unittest.TestCase):

    def setUp(self):

        # define the molecule
        path_hdf5 = (
            PATH_TEST / 'hdf5/C_adf_dzp.hdf5').absolute().as_posix()
        self.mol = Molecule(load=path_hdf5)

        # define the wave function
        self.wf = SlaterJastrow(self.mol, include_all_mo=True)

        # define the grid points
        self.npts = 21
        pts = get_pts(self.npts)

        self.pos = 10 * torch.ones(self.npts ** 2, self.mol.nelec * 3)
        self.pos[:, :3] = pts
        self.pos = Variable(self.pos)
        self.pos.requires_grad = True

    def test_mo(self):

        movals = self.wf.mo_scf(self.wf.ao(self.pos)).detach().numpy()

        for iorb in range(self.mol.basis.nmo):
            path_cube = PATH_TEST / f'cube/C_MO_%SCF_A%{iorb + 1}.cub'
            fname = path_cube.absolute().as_posix()
            adf_ref_data = np.array(read_cubefile(
                fname)).reshape(self.npts, self.npts)**2
            qmctorch_data = (movals[:, 0, iorb]).reshape(
                self.npts, self.npts)**2

            delta = np.abs(adf_ref_data - qmctorch_data)

            if __PLOT__:
                plt.subplot(1, 3, 1)
                plt.imshow(adf_ref_data)

                plt.subplot(1, 3, 2)
                plt.imshow(qmctorch_data)

                plt.subplot(1, 3, 3)
                plt.imshow(delta)
                plt.show()

            # the 0,0 point is much larger due to num instabilities
            delta = np.sort(delta.flatten())
            delta = delta[:-1]
            assert(delta.mean() < 1E-3)


if __name__ == "__main__":
    # unittest.main()
    t = TestMOvaluesADF()
    t.setUp()
    t.test_mo()



import unittest

import numpy as np
import torch

from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow
from ...path_utils import PATH_TEST
from .second_derivative import second_derivative

import matplotlib.pyplot as plt


class TestRadialSlater(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(0)
        np.random.seed(0)

        path_hdf5 = (
            PATH_TEST / 'hdf5/CO2_adf_dzp.hdf5').absolute().as_posix()
        self.mol = Molecule(load=path_hdf5)

        # wave function
        self.wf = SlaterJastrow(self.mol, kinetic='jacobi',
                                configs='ground_state',
                                include_all_mo=False)

    def test_first_derivative_x(self):

        npts = 1000
        self.pos = torch.zeros(npts, self.mol.nelec * 3)
        self.pos[:, 0] = torch.linspace(-4, 4, npts)
        self.dx = self.pos[1, 0] - self.pos[0, 0]

        xyz, r = self.wf.ao._process_position(self.pos)
        R, dR = self.wf.ao.radial(r, self.wf.ao.bas_n,
                                  self.wf.ao.bas_exp,
                                  xyz=xyz,
                                  derivative=[0, 1],
                                  sum_grad=False)

        R = R.detach().numpy()
        dR = dR.detach().numpy()
        ielec = 0

        for iorb in range(7):
            r0 = R[:, ielec, iorb]
            dz_r0 = dR[:, ielec, iorb, 0]
            dz_r0_fd = np.gradient(r0, self.dx)
            delta = np.delete(
                np.abs(dz_r0 - dz_r0_fd), np.s_[450:550])

            # plt.plot(dz_r0)
            # plt.plot(dz_r0_fd)
            # plt.show()

            assert(np.all(delta < 1E-3))

    def test_first_derivative_y(self):

        npts = 1000
        self.pos = torch.zeros(npts, self.mol.nelec * 3)
        self.pos[:, 1] = torch.linspace(-4, 4, npts)
        self.dy = self.pos[1, 1] - self.pos[0, 1]

        xyz, r = self.wf.ao._process_position(self.pos)
        R, dR = self.wf.ao.radial(r, self.wf.ao.bas_n,
                                  self.wf.ao.bas_exp,
                                  xyz=xyz,
                                  derivative=[0, 1],
                                  sum_grad=False)

        R = R.detach().numpy()
        dR = dR.detach().numpy()
        ielec = 0

        for iorb in range(7):
            r0 = R[:, ielec, iorb]
            dz_r0 = dR[:, ielec, iorb, 1]
            dz_r0_fd = np.gradient(r0, self.dy)
            delta = np.delete(
                np.abs(dz_r0 - dz_r0_fd), np.s_[450:550])

            # plt.plot(dz_r0)
            # plt.plot(dz_r0_fd)
            # plt.show()

            assert(np.all(delta < 1E-3))

    def test_first_derivative_z(self):

        npts = 1000
        self.pos = torch.zeros(npts, self.mol.nelec * 3)
        self.pos[:, 2] = torch.linspace(-4, 4, npts)
        self.dz = self.pos[1, 2] - self.pos[0, 2]

        xyz, r = self.wf.ao._process_position(self.pos)
        R, dR = self.wf.ao.radial(r, self.wf.ao.bas_n,
                                  self.wf.ao.bas_exp,
                                  xyz=xyz,
                                  derivative=[0, 1],
                                  sum_grad=False)
        R = R.detach().numpy()
        dR = dR.detach().numpy()
        ielec = 0

        for iorb in range(7):

            r0 = R[:, ielec, iorb]
            dz_r0 = dR[:, ielec, iorb, 2]
            dz_r0_fd = np.gradient(r0, self.dz)
            delta = np.delete(
                np.abs(dz_r0 - dz_r0_fd), np.s_[450:550])

            # plt.plot(dz_r0)
            # plt.plot(dz_r0_fd)
            # plt.show()

            assert(np.all(delta < 1E-3))

    def test_laplacian(self, eps=1E-4):

        npts = 1000

        self.pos = torch.zeros(npts, self.mol.nelec * 3)
        self.pos[:, 2] = torch.linspace(-4, 4, npts)
        eps = self.pos[1, 2] - self.pos[0, 2]

        self.pos[:, 2] = torch.linspace(-4, 4, npts)

        self.pos[:, 3] = eps
        self.pos[:, 5] = torch.linspace(-4, 4, npts)

        self.pos[:, 6] = -eps
        self.pos[:, 8] = torch.linspace(-4, 4, npts)

        self.pos[:, 10] = eps
        self.pos[:, 11] = torch.linspace(-4, 4, npts)

        self.pos[:, 13] = -eps
        self.pos[:, 14] = torch.linspace(-4, 4, npts)

        xyz, r = self.wf.ao._process_position(self.pos)
        R, dR, d2R = self.wf.ao.radial(r, self.wf.ao.bas_n,
                                       self.wf.ao.bas_exp,
                                       xyz=xyz,
                                       derivative=[0, 1, 2],
                                       sum_grad=False)

        for iorb in range(7):

            lap_analytic = np.zeros(npts - 2)
            lap_fd = np.zeros(npts - 2)

            for i in range(1, npts - 1):
                lap_analytic[i-1] = d2R[i, 0, iorb]

                r0 = R[i, 0, iorb].detach().numpy()
                rpz = R[i+1, 0, iorb].detach().numpy()
                rmz = R[i-1, 0, iorb].detach().numpy()
                d2z = second_derivative(rmz, r0, rpz, eps)

                r0 = R[i, 0, iorb]
                rpx = R[i, 1, iorb]
                rmx = R[i, 2, iorb]
                d2x = second_derivative(rmx, r0, rpx, eps)

                r0 = R[i, 0, iorb]
                rpy = R[i, 3, iorb]
                rmy = R[i, 4, iorb]
                d2y = second_derivative(rmy, r0, rpy, eps)

                lap_fd[i-1] = d2x + d2y + d2z

            delta = np.delete(
                np.abs(lap_analytic - lap_fd), np.s_[450:550])

            assert(np.all(delta < 5E-3))

            # plt.plot(lap_analytic, linewidth=2)
            # plt.plot(lap_fd)
            # plt.show()


if __name__ == "__main__":
    unittest.main()

    # t = TestRadialSlater()
    # t.setUp()
    # t.test_first_derivative_x()
    # # t.test_first_derivative_y()
    # # t.test_first_derivative_z()
    # t.test_laplacian()



import torch


class OrbitalProjector:

    def __init__(self, configs, mol, cuda=False):
        """Project the MO matrix in Slater Matrices

        Args:
            configs (list): configurations of the slater determinants
            mol (Molecule): Molecule object
            cuda (bool): use cuda or not
        """

        self.configs = configs
        self.nconfs = len(configs[0])
        self.nmo = mol.basis.nmo
        self.nup = mol.nup
        self.ndown = mol.ndown
        self.device = torch.device('cpu')
        if cuda:
            self.device = torch.device('cuda')

    def get_projectors(self):
        """Get the projectors of the conf in the CI expansion

        Returns:
            torch.tensor, torch.tensor : projectors
        """

        Pup = torch.zeros(self.nconfs, self.nmo, self.nup)
        Pdown = torch.zeros(self.nconfs, self.nmo, self.ndown)

        for ic, (cup, cdown) in enumerate(
                zip(self.configs[0], self.configs[1])):

            for _id, imo in enumerate(cup):
                Pup[ic][imo, _id] = 1.

            for _id, imo in enumerate(cdown):
                Pdown[ic][imo, _id] = 1.

        return Pup.unsqueeze(1).to(self.device), Pdown.unsqueeze(1).to(self.device)

    def split_orbitals(self, mat):
        """Split the orbital  matrix in multiple slater matrices

        Args:
            mat (torch.tensor): matrix to split

        Returns:
            torch.tensor: all slater matrices
        """
        if not hasattr(self, 'Pup'):
            self.Pup, self.Pdown = self.get_projectors()

        if mat.ndim == 4:
            # case for multiple operators
            out_up = mat[..., :self.nup, :] @ self.Pup.unsqueeze(1)
            out_down = mat[..., self.nup:,
                           :] @ self.Pdown.unsqueeze(1)

        else:
            # case for single operator
            out_up = mat[..., :self.nup, :] @ self.Pup
            out_down = mat[..., self.nup:, :] @ self.Pdown

        return out_up, out_down


class ExcitationMask:

    def __init__(self, unique_excitations, mol, max_orb, cuda=False):
        """Select the occupied MOs of Slater determinant using masks

        Args:
            unique_excitations (list): the list of unique excitations
            mol (Molecule): Molecule object
            max_orb (list): the max index of each orb for each spin
            cuda (bool): use cuda or not
        """

        self.unique_excitations = unique_excitations
        self.num_unique_exc = len(unique_excitations[0])
        self.nmo = mol.basis.nmo
        self.nup = mol.nup
        self.ndown = mol.ndown
        self.nelec = mol.nelec
        self.max_orb = max_orb

        self.device = torch.device('cpu')
        if cuda:
            self.device = torch.device('cuda')

    def get_index_unique_single(self):
        """Computes the 1D index and permutation
           for the unique singles."""

        ncol_up = self.max_orb[0]-self.nup
        ncol_down = self.max_orb[1]-self.ndown

        self.index_unique_single_up = []
        self.index_unique_single_down = []

        self.sign_unique_single_up = []
        self.sign_unique_single_down = []

        for exc_up, exc_down in zip(self.unique_excitations[0],
                                    self.unique_excitations[1]):

            if len(exc_up[0]) == 1:
                ielec, iorb = exc_up[0][0], exc_up[1][0]
                icol = iorb-self.nup

                self.index_unique_single_up.append(
                    ielec*ncol_up + icol)

                npermut = self.nup-ielec-1
                self.sign_unique_single_up.append((-1)**(npermut))

            if len(exc_down[1]) == 1:
                ielec, iorb = exc_down[0][0], exc_down[1][0]
                icol = iorb-self.ndown

                self.index_unique_single_down.append(
                    ielec*ncol_down + icol)

                npermut = self.ndown-ielec-1
                self.sign_unique_single_down.append((-1)**(npermut))

        self.sign_unique_single_up = torch.as_tensor(
            self.sign_unique_single_up).to(self.device)
        self.sign_unique_single_down = torch.as_tensor(
            self.sign_unique_single_down).to(self.device)

    def get_index_unique_double(self):
        """Computes the 1D index of the double excitation matrices."""

        ncol_up = self.max_orb[0]-self.nup
        ncol_down = self.max_orb[1]-self.ndown

        self.index_unique_double_up = []
        self.index_unique_double_down = []

        for exc_up, exc_down in zip(self.unique_excitations[0],
                                    self.unique_excitations[1]):

            if len(exc_up[0]) == 2:
                for ielec in exc_up[0]:
                    for iorb in exc_up[1]:
                        icol = iorb-self.nup
                        self.index_unique_double_up.append(
                            ielec*ncol_up + icol)

            if len(exc_down[1]) == 2:
                for ielec in exc_up[0]:
                    for iorb in exc_up[1]:
                        icol = iorb-self.ndown
                        self.index_unique_double_down.append(
                            ielec*ncol_down + icol)



from .fully_connected_jastrow_kernel import FullyConnectedJastrowKernel
from .jastrow_kernel_electron_electron_base import JastrowKernelElectronElectronBase
from .pade_jastrow_kernel import PadeJastrowKernel
from .pade_jastrow_polynomial_kernel import PadeJastrowPolynomialKernel





import numpy as np
import torch
from .slater_jastrow import SlaterJastrow

import operator

from .jastrows.elec_elec.kernels.pade_jastrow_kernel import PadeJastrowKernel as PadeJastrowKernelElecElec
from .jastrows.jastrow_factor_combined_terms import JastrowFactorCombinedTerms
from .jastrows.elec_nuclei.kernels.pade_jastrow_kernel import PadeJastrowKernel as PadeJastrowKernelElecNuc


from .orbitals.atomic_orbitals_backflow import AtomicOrbitalsBackFlow
from .orbitals.atomic_orbitals_orbital_dependent_backflow import AtomicOrbitalsOrbitalDependentBackFlow
from .orbitals.backflow.kernels import BackFlowKernelInverse


class SlaterManyBodyJastrowBackflow(SlaterJastrow):

    def __init__(self, mol, configs='ground_state',
                 kinetic='jacobi',
                 jastrow_kernel={
                     'ee': PadeJastrowKernelElecElec,
                     'en': PadeJastrowKernelElecNuc,
                     'een': None},
                 jastrow_kernel_kwargs={
                     'ee': {},
                     'en': {},
                     'een': {}},
                 backflow_kernel=BackFlowKernelInverse,
                 backflow_kernel_kwargs={},
                 orbital_dependent_backflow=False,
                 cuda=False,
                 include_all_mo=True):
        """Slater Jastrow wave function with many-body Jastrow factor and backflow

        .. math::
            \\Psi(R_{at}, r) = J(R_{at}, r)\\sum_n c_n D^\\uparrow_n(q^\\uparrow)D^\\downarrow_n(q^\\downarrow)

        with

        .. math::
            J(r) = \\exp\\left( K_{ee}(r) + K_{en}(R_{at},r) + K_{een}(R_{at}, r) \\right)    

        with the different kernels representing electron-electron, electron-nuclei and electron-electron-nuclei terms and

        .. math::
            q(r_i) = r_i + \\sum){j\\neq i} K_{BF}(r_{ij})(r_i-r_j)

        is a backflow transformation defined by the kernel K_{BF}. Note that different transformation
        can be used for different orbital via the `orbital_dependent_backflow` option.

        Args:
        Args:
            mol (Molecule): a QMCTorch molecule object
            configs (str, optional): defines the CI configurations to be used. Defaults to 'ground_state'.
                - ground_state : only the ground state determinant in the wave function
                - single(n,m) : only single excitation with n electrons and m orbitals 
                - single_double(n,m) : single and double excitation with n electrons and m orbitals
                - cas(n, m) : all possible configuration using n eletrons and m orbitals                   
            kinetic (str, optional): method to compute the kinetic energy. Defaults to 'jacobi'.
                - jacobi : use the Jacobi formula to compute the kinetic energy 
                - auto : use automatic differentiation to compute the kinetic energy
            jastrow_kernel (dict, optional) : different Jastrow kernels for the different terms. 
                By default only electron-electron and electron-nuclei terms are used
            jastrow_kernel_kwargs (dict, optional) : keyword arguments for the jastrow kernels contructor
            backflow_kernel (BackFlowKernelBase, optional) : kernel function of the backflow transformation. 
                - By default an inverse kernel K(r_{ij}) = w/r_{ij} is used
            backflow_kernel_kwargs (dict, optional) : keyword arguments for the backflow kernel contructor
            orbital_dependent_backflow (bool, optional) : every orbital has a different transformation if True. Default to False 
            cuda (bool, optional): turns GPU ON/OFF  Defaults to False.
            include_all_mo (bool, optional): include either all molecular orbitals or only the ones that are
                                             popualted in the configs. Defaults to False
    
        Examples::
            >>> from qmctorch.scf import Molecule
            >>> from qmctorch.wavefunction import SlaterManyBodyJastrowBackflow
            >>> mol = Molecule('h2o.xyz', calculator='adf', basis = 'dzp')
            >>> wf = SlaterManyBodyJastrowBackflow(mol, configs='cas(2,2)')
        """

        super().__init__(mol, configs, kinetic, None, {}, cuda, include_all_mo)

        # process the backflow transformation
        if orbital_dependent_backflow:
            self.ao = AtomicOrbitalsOrbitalDependentBackFlow(
                mol, backflow_kernel, backflow_kernel_kwargs, cuda)
        else:
            self.ao = AtomicOrbitalsBackFlow(
                mol, backflow_kernel, backflow_kernel_kwargs, cuda)

        if self.cuda:
            self.ao = self.ao.to(self.device)

        # process the Jastrow
        if jastrow_kernel is not None:

            for k in ['ee', 'en', 'een']:
                if k not in jastrow_kernel.keys():
                    jastrow_kernel[k] = None
                if k not in jastrow_kernel_kwargs.keys():
                    jastrow_kernel_kwargs[k] = None

            self.use_jastrow = True
            self.jastrow_type = 'JastrowFactorCombinedTerms'

            self.jastrow = JastrowFactorCombinedTerms(
                self.mol.nup, self.mol.ndown,
                torch.as_tensor(self.mol.atom_coords),
                jastrow_kernel=jastrow_kernel,
                jastrow_kernel_kwargs=jastrow_kernel_kwargs,
                cuda=cuda)

            if self.cuda:
                for term in self.jastrow.jastrow_terms:
                    term = term.to(self.device)

        self.log_data()

    def forward(self, x, ao=None):
        """computes the value of the wave function for the sampling points

        .. math::
            J(R) \\Psi(R) =  J(R) \\sum_{n} c_n D^{u}_n(r^u) \\times D^{d}_n(r^d)

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)
            ao (torch.tensor, optional): values of the atomic orbitals (Nbatch, Nelec, Nao)

        Returns:
            torch.tensor: values of the wave functions at each sampling point (Nbatch, 1)

        Examples::
            >>> mol = Molecule('h2.xyz', calculator='adf', basis = 'dzp')
            >>> wf = SlaterJastrow(mol, configs='cas(2,2)')
            >>> pos = torch.rand(500,6)
            >>> vals = wf(pos)
        """

        # compute the jastrow from the pos
        if self.use_jastrow:
            J = self.jastrow(x)

        # atomic orbital
        if ao is None:
            x = self.ao(x)
        else:
            x = ao

        # molecular orbitals
        x = self.mo_scf(x)

        # mix the mos
        x = self.mo(x)

        # pool the mos
        x = self.pool(x)

        # compute the CI and return
        if self.use_jastrow:
            return J * self.fc(x)

        else:
            return self.fc(x)

    def ao2mo(self, ao):
        """transforms AO values in to MO values."""
        return self.mo(self.mo_scf(ao))

    def pos2mo(self, x, derivative=0, sum_grad=True):
        """Compute the MO vals from the pos

        Args:
            x ([type]): [description]
            derivative (int, optional): [description]. Defaults to 0.
            sum_grad (bool, optional): [description]. Defaults to True.

        Returns:
            [type]: [description]
        """

        ao = self.ao(x, derivative=derivative, sum_grad=sum_grad)
        return self.ao2mo(ao)

    def kinetic_energy_jacobi(self, x,  **kwargs):
        r"""Compute the value of the kinetic enery using the Jacobi Formula.


        .. math::
             \\frac{\Delta (J(R) \Psi(R))}{ J(R) \Psi(R)} = \\frac{\\Delta J(R)}{J(R}
                                                          + 2 \\frac{\\nabla J(R)}{J(R)} \\frac{\\nabla \\Psi(R)}{\\Psi(R)}
                                                          + \\frac{\\Delta \\Psi(R)}{\\Psi(R)}

        The lapacian of the determinental part is computed via

        .. math::
            \\Delta_i \\Psi(R) \\sum_n c_n ( \\frac{\\Delta_i D_n^{u}}{D_n^{u}} +
                                           \\frac{\\Delta_i D_n^{d}}{D_n^{d}} +
                                           2 \\frac{\\nabla_i D_n^{u}}{D_n^{u}} \\frac{\\nabla_i D_n^{d}}{D_n^{d}} )
                                           D_n^{u} D_n^{d}

        Since the backflow orbitals are multi-electronic the laplacian of the determinants
        are obtained

        .. math::
            \\frac{\\Delta det(A)}{det(A)} = Tr(A^{-1} \\Delta A) +
                                             Tr(A^{-1} \\nabla A) Tr(A^{-1} \\nabla A) +
                                             Tr( (A^{-1} \\nabla A) (A^{-1} \\nabla A ))


        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            torch.tensor: values of the kinetic energy at each sampling points
        """

        # get ao values
        ao, dao, d2ao = self.ao(
            x, derivative=[0, 1, 2], sum_grad=False)

        # get the mo values
        mo = self.ao2mo(ao)
        dmo = self.ao2mo(dao)
        d2mo = self.ao2mo(d2ao)

        # compute the value of the slater det
        slater_dets = self.pool(mo)
        sum_slater_dets = self.fc(slater_dets)

        # compute ( tr(A_u^-1\Delta A_u) + tr(A_d^-1\Delta A_d) )
        hess = self.pool.operator(mo, d2mo)

        # compute (tr(A_u^-1\nabla A_u) and tr(A_d^-1\nabla A_d))
        grad = self.pool.operator(mo, dmo, op=None)

        # compute (tr((A_u^-1\nabla A_u)^2) + tr((A_d^-1\nabla A_d))^2)
        grad2 = self.pool.operator(mo, dmo, op_squared=True)

        # assemble the total second derivative term
        hess = (hess.sum(0)
                + operator.add(*[(g**2).sum(0) for g in grad])
                - grad2.sum(0)
                + 2 * operator.mul(*grad).sum(0))

        hess = self.fc(hess * slater_dets) / sum_slater_dets

        if self.use_jastrow is False:
            return -0.5 * hess

        # compute the Jastrow terms
        jast, djast, d2jast = self.jastrow(x,
                                           derivative=[0, 1, 2],
                                           sum_grad=False)

        # prepare the second derivative term d2Jast/Jast
        # Nbatch x Nelec
        d2jast = d2jast / jast

        # prepare the first derivative term
        djast = djast / jast.unsqueeze(-1)

        # -> Nelec x Ndim x Nbatch
        djast = djast.permute(2, 1, 0)

        # -> [Nelec*Ndim] x Nbatch
        djast = djast.reshape(-1, djast.shape[-1])

        # prepare the grad of the dets
        # [Nelec*Ndim] x Nbatch x 1
        grad_val = self.fc(operator.add(*grad) *
                           slater_dets) / sum_slater_dets

        # [Nelec*Ndim] x Nbatch
        grad_val = grad_val.squeeze()

        # assemble the derivaite terms
        out = d2jast.sum(-1) + 2*(grad_val * djast).sum(0) + \
            hess.squeeze(-1)

        return -0.5 * out.unsqueeze(-1)

    def gradients_jacobi(self, x, sum_grad=True):
        """Computes the gradients of the wf using Jacobi's Formula

        Args:
            x ([type]): [description]
        """
        raise NotImplementedError(
            'Gradient through Jacobi formulat not implemented for backflow orbitals')



import unittest

import numpy as np
import torch
import torch.optim as optim

from qmctorch.sampler import Metropolis
from qmctorch.solver import Solver
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow


class TestLiH(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(0)
        np.random.seed(0)

        # molecule
        self.mol = Molecule(
            atom='Li 0 0 0; H 0 0 3.015',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        # wave function
        self.wf = SlaterJastrow(self.mol, kinetic='jacobi',
                                configs='single(2,2)',
                                include_all_mo=False)

        # sampler
        self.sampler = Metropolis(
            nwalkers=500,
            nstep=200,
            step_size=0.05,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        # optimizer
        self.opt = optim.Adam(self.wf.parameters(), lr=0.01)

        # solver
        self.solver = Solver(wf=self.wf, sampler=self.sampler,
                                          optimizer=self.opt)

    def test1_single_point(self):

        # sample and compute observables
        obs = self.solver.single_point()
        e, v = obs.energy, obs.variance

        # # values on different arch
        # expected_energy = [-1.1464850902557373,
        #                    -1.14937478612449]

        # # values on different arch
        # expected_variance = [0.9279592633247375,
        #                      0.7445300449383236]

        # assert(np.any(np.isclose(e.data.item(), np.array(expected_energy))))
        # assert(np.any(np.isclose(v.data.item(), np.array(expected_variance))))

    def test2_wf_opt_grad_auto(self):
        self.solver.sampler = self.sampler

        self.solver.configure(track=['local_energy'],
                              loss='energy', grad='auto')
        obs = self.solver.run(5)

    def test3_wf_opt_grad_manual(self):
        self.solver.sampler = self.sampler

        self.solver.configure(track=['local_energy'],
                              loss='energy', grad='manual')
        obs = self.solver.run(5)


if __name__ == "__main__":
    unittest.main()
    # t = TestLiH()
    # t.setUp()
    # t.test1_single_point()
    # t.test2_wf_opt_grad_auto()
    # t.test3_wf_opt_grad_manual()



__all__ = ['WaveFunction', 'SlaterJastrow', 'SlaterManyBodyJastrow',
           'SlaterJastrowBackFlow', 'SlaterOrbitalDependentJastrow',
           'SlaterManyBodyJastrowBackflow']

from .wf_base import WaveFunction
from .slater_jastrow import SlaterJastrow
from .slater_combined_jastrow import SlaterManyBodyJastrow
from .slater_jastrow_backflow import SlaterJastrowBackFlow
from .slater_combined_jastrow_backflow import SlaterManyBodyJastrowBackflow
from .slater_orbital_dependent_jastrow import SlaterOrbitalDependentJastrow



import unittest

import torch
from pyscf import gto
from torch.autograd import Variable, grad, gradcheck
import numpy as np
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow
from qmctorch.wavefunction.orbitals.atomic_orbitals_backflow import AtomicOrbitalsBackFlow
from qmctorch.wavefunction.orbitals.backflow.kernels import BackFlowKernelInverse
torch.set_default_tensor_type(torch.DoubleTensor)

torch.manual_seed(101)
np.random.seed(101)


def hess(out, pos):

    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)

    for idim in range(jacob.shape[1]):

        tmp = grad(jacob[:, idim], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, idim] = tmp[:, idim]

    return hess


def hess_single_element(out, inp):

    shape = out.shape
    out = out.reshape(-1, 1)

    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, inp,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape))

    hess = grad(jacob, inp,
                grad_outputs=z,
                only_inputs=True,
                create_graph=True)[0]

    return hess.reshape(*shape)


class TestBFAOderivativesPyscf(unittest.TestCase):

    def setUp(self):

        # define the molecule
        at = 'C 0 0 0'
        basis = 'dzp'
        self.mol = Molecule(atom=at,
                            calculator='pyscf',
                            basis=basis,
                            unit='bohr')

        # define the wave function
        self.ao = AtomicOrbitalsBackFlow(
            self.mol, BackFlowKernelInverse)

        # define the grid points
        self.npts = 11
        self.pos = torch.rand(self.npts, self.mol.nelec * 3)
        self.pos = Variable(self.pos)
        self.pos.requires_grad = True

    def test_ao_gradian(self):
        """Test the calculation of the gradient of the at
        wrt the original coordinates."""

        ao = self.ao(self.pos)
        dao = self.ao(self.pos, derivative=1, sum_grad=False)

        dao_grad = grad(
            ao, self.pos, grad_outputs=torch.ones_like(ao))[0]

        assert(torch.allclose(dao.sum(), dao_grad.sum()))

        dao = dao.sum(-1).sum(-1)
        dao_grad = dao_grad.T
        assert(torch.allclose(dao, dao_grad))

    def test_ao_jacobian(self):

        ao = self.ao(self.pos)
        dao = self.ao(self.pos, derivative=1)

        dao_grad = grad(
            ao, self.pos, grad_outputs=torch.ones_like(ao))[0]

        assert(torch.allclose(dao.sum(), dao_grad.sum()))

        dao = dao.sum(-1).sum(-1)
        dao_grad = dao_grad.reshape(-1, self.ao.nelec, 3).sum(-1)
        dao_grad = dao_grad.T
        assert(torch.allclose(dao, dao_grad))

    def test_ao_hess(self):

        ao = self.ao(self.pos)
        d2ao = self.ao(self.pos, derivative=2)

        d2ao_grad = hess(ao, self.pos)
        assert(torch.allclose(d2ao.sum(), d2ao_grad.sum()))

        d2ao = d2ao.sum(-1).sum(-1)
        d2ao_grad = d2ao_grad.reshape(-1, self.ao.nelec, 3).sum(-1)
        d2ao_grad = d2ao_grad.T
        assert(torch.allclose(d2ao, d2ao_grad))

    def test_all_ao_values(self):
        ao = self.ao(self.pos)
        dao = self.ao(self.pos, derivative=1, sum_grad=False)
        d2ao = self.ao(self.pos, derivative=2, sum_hess=False)
        ao_all, dao_all, d2ao_all = self.ao(
            self.pos, derivative=[0, 1, 2])

        assert(torch.allclose(ao, ao_all))
        assert(torch.allclose(dao, dao_all))
        assert(torch.allclose(d2ao, d2ao_all))


if __name__ == "__main__":
    unittest.main()



import torch
from torch import nn


class OrbitalDependentBackFlowKernel(nn.Module):

    def __init__(self, backflow_kernel, backflow_kernel_kwargs, mol, cuda):
        """Compute orbital dependent back flow kernel, i.e. the functions
        f(rij) where rij is the distance between electron i and j
        This kernel is used in the backflow transformation
        .. math:
            q^{\\alpha}_i = r_i + \\sum_{j\\neq i} f^{\\alpha}(r_{ij}) (r_i-r_j)

        where :math: `f^{\\alpha}(r_{ij})` is the kernel for obital :math: `\\alpha`
        """
        super().__init__()

        self.nelec = mol.nelec
        self.nao = mol.basis.nao
        self.orbital_dependent_kernel = nn.ModuleList(
            [backflow_kernel(mol, cuda, **backflow_kernel_kwargs) for iao in range(self.nao)])

        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')

        # domension along which the different orbitals are stacked
        # with stach_axis = 1 the resulting tensors will have dimension
        # Nbatch x Nao x ...
        self.stack_axis = 1

    def forward(self, ree, derivative=0):
        """Computes the desired values of the kernels
         Args:
            ree (torch.tensor): e-e distance Nbatch x Nelec x Nelec
            derivative (int): derivative requried 0, 1, 2

        Returns:
            torch.tensor : f(r) Nbatch x Nao x Nelec x Nelec
        """
        out = None

        for ker in self.orbital_dependent_kernel:
            ker_val = ker(ree, derivative).unsqueeze(self.stack_axis)
            if out is None:
                out = ker_val
            else:
                out = torch.cat((out, ker_val), axis=self.stack_axis)
        return out



import unittest

import numpy as np
import torch
from torch.autograd import grad, Variable

from qmctorch.wavefunction.orbitals.spherical_harmonics import Harmonics


def hess(out, pos):

    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)

    for idim in range(jacob.shape[1]):

        tmp = grad(jacob[:, idim], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, idim] = tmp[:, idim]

    return hess


def hess_mixed_terms(out, pos):

    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)
    nelec = pos.shape[1]//3
    k = 0

    for ielec in range(nelec):

        ix = ielec*3
        tmp = grad(jacob[:, ix], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, k] = tmp[:, ix+1]
        k = k + 1
        hess[:, k] = tmp[:, ix+2]
        k = k + 1

        iy = ielec*3 + 1
        tmp = grad(jacob[:, iy], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, k] = tmp[:, iy+1]
        k = k + 1

    return hess


class TestCartesianHarmonics(unittest.TestCase):

    def setUp(self):
        bas_kx = torch.as_tensor([0, 1, 0, 0, 2, 0, 0, 1, 0, 0, 1])
        bas_ky = torch.as_tensor([0, 0, 1, 0, 0, 2, 0, 1, 1, 0, 1])
        bas_kz = torch.as_tensor([0, 0, 0, 1, 0, 0, 2, 0, 1, 1, 1])
        self.nbas = len(bas_kx)

        self.harmonics = Harmonics(
            'cart', bas_kx=bas_kx, bas_ky=bas_ky, bas_kz=bas_kz)

        self.nbatch = 10
        self.nelec = 4

        self.pos = Variable(torch.rand(self.nbatch, self.nelec*3))
        self.pos.requires_grad = True

    def process_position(self):
        """Return the distance between electron and centers."""

        bas_coords = torch.zeros(self.nbas, 3)
        xyz = (self.pos.view(-1, self.nelec, 1, 3) -
               bas_coords[None, ...])
        r = torch.sqrt((xyz*xyz).sum(3))
        return xyz, r

    def test_value(self):
        xyz, r = self.process_position()
        self.harmonics(xyz, derivative=0)

    def test_grad(self):
        xyz, r = self.process_position()

        val_grad = self.harmonics(
            xyz, derivative=1, sum_grad=False)

        val = self.harmonics(xyz)
        val_grad_auto = grad(val, self.pos, torch.ones_like(val))[0]

        assert(torch.allclose(
            val_grad.sum(), val_grad_auto.sum(), atol=1E-6))

    def test_jac(self):
        xyz, r = self.process_position()
        val_jac = self.harmonics(xyz, derivative=1, sum_grad=True)
        val = self.harmonics(xyz)
        val_jac_auto = grad(val, self.pos, torch.ones_like(val))[0]

        assert(torch.allclose(
            val_jac.sum(), val_jac_auto.sum(), atol=1E-6))

    def test_lap(self):
        xyz, r = self.process_position()
        val_hess = self.harmonics(xyz, derivative=2)
        val = self.harmonics(xyz)
        val_hess_auto = hess(val, self.pos)

        assert(torch.allclose(
            val_hess.sum(), val_hess_auto.sum(), atol=1E-6))

    def test_mixed_der(self):
        xyz, r = self.process_position()
        val_hess = self.harmonics(xyz, derivative=3)
        val = self.harmonics(xyz)
        val_hess_auto = hess_mixed_terms(val, self.pos)

        assert(torch.allclose(
            val_hess.sum(), val_hess_auto.sum(), atol=1E-6))


if __name__ == "__main__":
    unittest.main()



from types import SimpleNamespace

import h5py
import numpy as np
import torch

from .. import log


def print_insert_error(obj, obj_name):
    print(obj_name, obj)
    log.critical('Issue inserting data {0} of type {type}',
                 obj_name, type=str(type(obj)))


def print_insert_type_error(obj, obj_name):
    log.critical('Issue inserting type of data {0}} ({type}})' %
                 obj_name, type=str(type(obj)))
    
def print_load_error(grp):
    log.critical('Issue loading {grp}', grp=grp)


def load_from_hdf5(obj, fname, obj_name):
    """Load the content of an hdf5 file in an object.

    Arguments:
        obj {object} -- object where to load the data
        fname {str} -- name pf the hdf5 file
        obj_name {str} -- name of the root group in the hdf5
    """

    h5 = h5py.File(fname, 'r')
    root_grp = h5[obj_name]

    load_object(root_grp, obj, obj_name)
    h5.close()


def load_object(grp, parent_obj, grp_name):
    """Load object attribute from the hdf5 group/data

    Arguments:
        grp {hdf5 group} -- the current group in the hdf5 architecture
        parent_obj {object} -- parent object
        grp_name {str} -- name of the group
    """

    for child_grp_name, child_grp in grp.items():

        if isgroup(child_grp):
            load_group(child_grp, parent_obj, child_grp_name)
        else:
            load_data(child_grp, parent_obj, child_grp_name)


def load_group(grp, parent_obj, grp_name):
    """Load object attribute from the hdf5 group

    Arguments:
        grp {hdf5 group} -- the current group in the hdf5 architecture
        parent_obj {object} -- parent object
        grp_name {str} -- name of the group
    """
    try:
        if not hasattr(parent_obj, grp_name):
            parent_obj.__setattr__(
                grp_name, SimpleNamespace())
        load_object(grp,
                    parent_obj.__getattribute__(
                        grp_name),
                    grp_name)
    except:
        print_load_error(grp_name)


def load_data(grp, parent_obj, grp_name):
    """Load data from the hdf5 data

    Arguments:
        grp {hdf5 group} -- the current group in the hdf5 architecture
        parent_obj {object} -- parent object
        grp_name {str} -- name of the group
    """
    try:
        parent_obj.__setattr__(grp_name,
                               cast_loaded_data(grp[()]))
    except:
        print_load_error(grp_name)


def cast_loaded_data(data):
    """cast the data before loading."""

    cast_fn = {bytes: bytes2str}

    if type(data) in cast_fn:
        data = cast_fn[type(data)](data)

    return data


def bytes2str(bstr):
    """Convert a bytes into string."""
    if type(bstr) is bytes:
        return bstr.decode('utf-8')
    elif type(bstr) is str:
        return bstr
    else:
        raise TypeError(
            bstr, ' should be a bytes or str but got ', type(bstr), ' instead')


def lookup_cast(ori_type, current_type):
    raise NotImplementedError(
        "cast the data to the type contained in .attrs['type']")


def isgroup(grp):
    """Check if current hdf5 group is a group

    Arguments:
        grp {hdf5 group} -- hdf5 group or dataset

    Returns:
        bool -- True if the group is a group
    """

    return type(grp) == h5py._hl.group.Group or type(grp) == h5py._hl.files.File


def dump_to_hdf5(obj, fname, root_name=None):
    """Dump the content of an object in a hdf5 file.

    Arguments:
        obj {object} -- object to dump
        fname {str} -- name of the hdf5

    Keyword Arguments:
        root_name {str} -- root group in the hdf5 file (default: {None})
    """

    h5 = h5py.File(fname, 'a')

    if root_name is None:
        root_name = obj.__class__.__name__

    # change root name if that name is already present in the file
    if root_name in h5:

        log.info('')
        log.info(' Warning : dump to hdf5')
        log.info(
            ' Object {obj} already exists in {parent}', obj=root_name, parent=fname)
        n = sum(1 for n in h5 if n.startswith(root_name)) + 1
        root_name = root_name + '_' + str(n)
        log.info(
            ' Object name changed to {obj}', obj=root_name)
        log.info('')

    insert_object(obj, h5, root_name)
    h5.close()
    return root_name


def insert_object(obj, parent_grp, obj_name):
    """Insert the content of the object in the hdf5 file

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """

    if haschildren(obj):
        insert_group(obj, parent_grp, obj_name)
    else:
        insert_data(obj, parent_grp, obj_name)


def insert_group(obj, parent_grp, obj_name):
    """Insert the content of the object in a hdf5 group

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """

    # ignore object starting with underscore
    # a lot of pytorch internal are like that
    if obj_name.startswith('_'):
        log.debug(
            ' Warning : Object {obj} not stored in {parent}', obj=obj_name, parent=parent_grp)
        log.debug(
            '         : because object name starts with "_"')
        return

    # store if the object name is not in parent
    if obj_name not in parent_grp:

        try:
            own_grp = parent_grp.create_group(obj_name)

            for child_name in get_children_names(obj):
                child_obj = get_child_object(obj, child_name)
                insert_object(child_obj,  own_grp, child_name)

        except Exception as inst:
            print(type(inst))
            print(inst)
            
            print_insert_error(obj, obj_name)

    # if something went wrong anyway
    else:
        log.critical(
            ' Warning : Object {obj} already exists in {parent}', obj=obj_name, parent=parent_grp)
        log.critical(
            ' Warning : Keeping original version of the data')


def insert_data(obj, parent_grp, obj_name):
    """Insert the content of the object in a hdf5 dataset

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """

    if obj_name.startswith('_'):
        return

    try:
        lookup_insert = {list: insert_list,
                         tuple: insert_tuple,
                         np.ndarray: insert_numpy,
                         torch.Tensor: insert_torch_tensor,
                         torch.nn.parameter.Parameter: insert_torch_parameter,
                         torch.device: insert_none,
                         type(None): insert_none}

        insert_fn = lookup_insert[type(obj)]
    except KeyError:
        insert_fn = insert_default

    try:
        insert_fn(obj, parent_grp, obj_name)
        # insert_type(obj, parent_grp, obj_name)
    except Exception as expt_message:
        print(expt_message)
        print_insert_error(obj, obj_name)


def insert_type(obj, parent_grp, obj_name):
    """Insert the content of the type object in an attribute

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """
    try:
        parent_grp[obj_name].attrs['type'] = str(type(obj))
    except:
        print_insert_type_error(obj, obj_name)


def insert_default(obj, parent_grp, obj_name):
    """Default funtion to insert a dataset

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """
    try:
        parent_grp.create_dataset(obj_name, data=obj)
    except Exception as expt_message:
        print(expt_message)
        print_insert_error(obj, obj_name)


def insert_list(obj, parent_grp, obj_name):
    """funtion to insert a list as a dataset

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """

    
    try:
        if np.all([isinstance(el,torch.Tensor) for el in obj]):
            obj = [el.numpy() for el in obj]
            
        parent_grp.create_dataset(obj_name, data=np.array(obj))
    except:
        for il, l in enumerate(obj):
            try:
                insert_object(l, parent_grp, obj_name+'_'+str(il))
            except:
                print_insert_error(obj, obj_name)


def insert_tuple(obj, parent_grp, obj_name):
    """funtion to insert a tuple as a dataset

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """
    insert_list(list(obj), parent_grp, obj_name)


def insert_numpy(obj, parent_grp, obj_name):
    """funtion to insert a numpy array as a dataset

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """
    if obj.dtype.str.startswith('<U'):
        obj = obj.astype('S')
    insert_default(obj, parent_grp, obj_name)


def insert_torch_tensor(obj, parent_grp, obj_name):
    """funtion to insert a torch tensor as a dataset

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """
    insert_numpy(obj.cpu().detach().numpy(), parent_grp, obj_name)


def insert_torch_parameter(obj, parent_grp, obj_name):
    """funtion to insert a torch parameter as a dataset

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """
    insert_torch_tensor(obj.data.cpu(), parent_grp, obj_name)


def insert_none(obj, parent_grp, obj_name):
    """funtion to insert a None Type as a dataset

    Arguments:
        obj {object} -- object to save
        parent_grp {hdf5 group} -- group where to dump
        obj_name {str} -- name of the object
    """
    return


def haschildren(obj):
    """Check if the object has children

    Arguments:
        obj {object} -- the object to check

    Returns:
        bool -- True if the object has children
    """
    ommit_type = [torch.nn.parameter.Parameter, torch.Tensor]
    if type(obj) in ommit_type:
        return False
    else:
        return hasattr(obj, '__dict__') or hasattr(obj, 'keys')


def children(obj):
    """Returns the children of the object as items

    Arguments:
        obj {object} -- the object to check

    Returns:
        dict -- items
    """

    if hasattr(obj, '__dict__'):
        return obj.__dict__.items()

    elif hasattr(obj, 'keys'):
        return obj.items()


def get_children_names(obj):
    """Returns the children names of the object as items

    Arguments:
        obj {object} -- the object to check

    Returns:
        dict -- items
    """

    if hasattr(obj, '__dict__'):
        names = list(obj.__dict__.keys())

    elif hasattr(obj, 'keys'):
        names = list(obj.keys())

    if hasattr(obj, '__extra_attr__'):
        names += obj.__extra_attr__

    if hasattr(obj, 'state_dict'):
        names += list(obj.state_dict().keys())

    return list(set(names))


def get_child_object(obj, child_name):
    """Return the child object

    Arguments:
        obj {object} -- parent object
        child_name {str} -- cild name

    Returns:
        object -- child object
    """

    if hasattr(obj, '__getattr__'):
        try:
            return obj.__getattr__(child_name)

        except AttributeError:
            pass

    if hasattr(obj, '__getattribute__'):
        try:
            return obj.__getattribute__(child_name)

        except AttributeError:
            pass

    if hasattr(obj, '__getitem__'):
        try:
            return obj.__getitem__(child_name)

        except AttributeError:
            pass


def add_group_attr(filename, grp_name, attr):
    """Add attribute to a given group

    Arguments:
        filename {str} -- name of the file
        grp_name {str} -- name of the group
        attr {dict} -- attrivutes to add
    """

    h5 = h5py.File(filename, 'a')
    for k, v in attr.items():
        h5[grp_name].attrs[k] = v
    h5.close()


def register_extra_attributes(obj, attr_names):
    """Register extra attribute to be able to dump them

    Arguments:
        obj {object} -- the object where we want to add attr
        attr_names {list} -- a list of attr names
    """
    if not hasattr(obj, '__extra_attr__'):
        obj.__extra_attr__ = []
    obj.__extra_attr__ += attr_names



import torch
from torch import nn
import operator as op

from ...utils import bdet2, btrace
from .orbital_configurations import get_excitation, get_unique_excitation
from .orbital_projector import ExcitationMask, OrbitalProjector


class SlaterPooling(nn.Module):

    """Applies a slater determinant pooling in the active space."""

    def __init__(self, config_method, configs, mol, cuda=False):
        """Computes the Sater determinants

        Args:
            config_method (str): method used to define the config
            configs (tuple): configuratin of the electrons
            mol (Molecule): Molecule instance
            cuda (bool, optional): Turns GPU ON/OFF. Defaults to False.

        """
        super(SlaterPooling, self).__init__()

        self.config_method = config_method

        self.configs = configs
        self.nconfs = len(configs[0])
        self.index_max_orb_up = self.configs[0].max().item() + 1
        self.index_max_orb_down = self.configs[1].max().item() + 1

        self.excitation_index = get_excitation(configs)
        self.unique_excitation, self.index_unique_excitation = get_unique_excitation(
            configs)

        self.nmo = mol.basis.nmo
        self.nup = mol.nup
        self.ndown = mol.ndown
        self.nelec = self.nup + self.ndown

        self.orb_proj = OrbitalProjector(configs, mol, cuda=cuda)
        self.exc_mask = ExcitationMask(self.unique_excitation, mol,
                                       (self.index_max_orb_up,
                                        self.index_max_orb_down),
                                       cuda=cuda)

        self.device = torch.device('cpu')
        if cuda:
            self.device = torch.device('cuda')

    def forward(self, input):
        """Computes the values of the determinats

        Args:
            input (torch.tensor): MO matrices nbatch x nelec x nmo

        Returns:
            torch.tensor: slater determinants
        """
        if self.config_method.startswith('cas('):
            return self.det_explicit(input)
        else:
            return self.det_single_double(input)

    def get_slater_matrices(self, input):
        """Computes the slater matrices

        Args:
            input (torch.tensor): MO matrices nbatch x nelec x nmo


        Returns:
            (torch.tensor, torch.tensor): slater matrices of spin up/down
        """
        return self.orb_proj.split_orbitals(input)

    def det_explicit(self, input):
        """Computes the values of the determinants from the slater matrices

        Args:
            input (torch.tensor): MO matrices nbatch x nelec x nmo

        Returns:
            torch.tensor: slater determinants
        """

        mo_up, mo_down = self.get_slater_matrices(input)
        return (torch.det(mo_up) * torch.det(mo_down)).transpose(0, 1)

    def det_single_double(self, input):
        """Computes the determinant of ground state + single + double

        Args:
            input (torch.tensor): MO matrices nbatch x nelec x nmo

        Returns:
            torch.tensor: slater determinants
        """

        # compute the determinant of the unique single excitation
        det_unique_up, det_unique_down = self.det_unique_single_double(
            input)

        # returns the product of spin up/down required by each excitation
        return (det_unique_up[:, self.index_unique_excitation[0]] *
                det_unique_down[:, self.index_unique_excitation[1]])

    def det_ground_state(self, input):
        """Computes the SD of the ground state

        Args:
            input (torch.tensor): MO matrices nbatch x nelec x nmo
        """

        return (torch.det(input[:, :self.nup, :self.nup]),
                torch.det(input[:, self.nup:, :self.ndown]))

    def det_unique_single_double(self, input):
        """Computes the SD of single/double excitations

        The determinants of the single excitations
        are calculated from the ground state determinant and
        the ground state Slater matrices whith one column modified.
        See : Monte Carlo Methods in ab initio quantum chemistry
        B.L. Hammond, appendix B1


        Note : if the state on coonfigs are specified in order
        we end up with excitations that comes from a deep orbital, the resulting
        slater matrix has one column changed (with the new orbital) and several
        permutation. We therefore need to multiply the slater determinant
        by (-1)^nperm.


        .. math::

            MO = [ A | B ]
            det(Exc_{ij}) = (det(A) * A^{-1} * B)_{i,j}

        Args:
            input (torch.tensor): MO matrices nbatch x nelec x nmo

        """

        nbatch = input.shape[0]

        if not hasattr(self.exc_mask, 'index_unique_single_up'):
            self.exc_mask.get_index_unique_single()

        if not hasattr(self.exc_mask, 'index_unique_double_up'):
            self.exc_mask.get_index_unique_double()

        do_single = len(self.exc_mask.index_unique_single_up) != 0
        do_double = len(self.exc_mask.index_unique_double_up) != 0

        # occupied orbital matrix + det and inv on spin up
        Aup = input[:, :self.nup, :self.nup]
        detAup = torch.det(Aup)

        # occupied orbital matrix + det and inv on spin down
        Adown = input[:, self.nup:, :self.ndown]
        detAdown = torch.det(Adown)

        # store all the dets we need
        det_out_up = detAup.unsqueeze(-1).clone()
        det_out_down = detAdown.unsqueeze(-1).clone()

        # return the ground state
        if self.config_method == 'ground_state':
            return det_out_up, det_out_down

        # inverse of the
        invAup = torch.inverse(Aup)
        invAdown = torch.inverse(Adown)

        # virtual orbital matrices spin up/down
        Bup = input[:, :self.nup, self.nup:self.index_max_orb_up]
        Bdown = input[:, self.nup:,
                      self.ndown: self.index_max_orb_down]

        # compute the products of Ain and B
        mat_exc_up = (invAup @ Bup)
        mat_exc_down = (invAdown @ Bdown)

        if do_single:

            # determinant of the unique excitation spin up
            det_single_up = mat_exc_up.view(
                nbatch, -1)[:, self.exc_mask.index_unique_single_up]

            # determinant of the unique excitation spin down
            det_single_down = mat_exc_down.view(
                nbatch, -1)[:, self.exc_mask.index_unique_single_down]

            # multiply with ground state determinant
            # and account for permutation for deep excitation
            det_single_up = detAup.unsqueeze(-1) * \
                det_single_up.view(nbatch, -1)

            # multiply with ground state determinant
            # and account for permutation for deep excitation
            det_single_down = detAdown.unsqueeze(-1) * \
                det_single_down.view(nbatch, -1)

            # accumulate the dets
            det_out_up = torch.cat((det_out_up, det_single_up), dim=1)
            det_out_down = torch.cat(
                (det_out_down, det_single_down), dim=1)

        if do_double:

            # det of unique spin up double exc
            det_double_up = mat_exc_up.view(
                nbatch, -1)[:, self.exc_mask.index_unique_double_up]

            det_double_up = bdet2(
                det_double_up.view(nbatch, -1, 2, 2))

            det_double_up = detAup.unsqueeze(-1) * det_double_up

            # det of unique spin down double exc
            det_double_down = mat_exc_down.view(
                nbatch, -1)[:, self.exc_mask.index_unique_double_down]

            det_double_down = bdet2(
                det_double_down.view(nbatch, -1, 2, 2))

            det_double_down = detAdown.unsqueeze(-1) * det_double_down

            det_out_up = torch.cat((det_out_up, det_double_up), dim=1)
            det_out_down = torch.cat(
                (det_out_down, det_double_down), dim=1)

        return det_out_up, det_out_down

    def operator(self, mo, bop, op=op.add, op_squared=False):
        """Computes the values of an opearator applied to the procuts of determinant

        Args:
            mo (torch.tensor): matrix of MO vals(Nbatch, Nelec, Nmo)
            bkin (torch.tensor): kinetic operator (Nbatch, Nelec, Nmo)
            op (operator): how to combine the up/down contribution
            op_squared (bool, optional) return the trace of the square of the product if True

        Returns:
            torch.tensor: kinetic energy
        """

        # get the values of the operator
        if self.config_method == 'ground_state':
            op_vals = self.operator_ground_state(mo, bop, op_squared)

        elif self.config_method.startswith('single'):
            op_vals = self.operator_single_double(mo, bop, op_squared)

        elif self.config_method.startswith('cas('):
            op_vals = self.operator_explicit(mo, bop, op_squared)

        else:
            raise ValueError(
                'Configuration %s not recognized' % self.config_method)

        # combine the values is necessary
        if op is not None:
            return op(*op_vals)
        else:
            return op_vals

    def operator_ground_state(self, mo, bop, op_squared=False):
        """Computes the values of any operator on gs only

        Args:
            mo (torch.tensor): matrix of molecular orbitals
            bkin (torch.tensor): matrix of kinetic operator
            op_squared (bool, optional) return the trace of the square of the product if True

        Returns:
            torch.tensor: operator values
        """

        # occupied orbital matrix + det and inv on spin up
        Aocc_up = mo[:, :self.nup, :self.nup]

        # occupied orbital matrix + det and inv on spin down
        Aocc_down = mo[:, self.nup:, :self.ndown]

        # inverse of the
        invAup = torch.inverse(Aocc_up)
        invAdown = torch.inverse(Aocc_down)

        # precompute the product A^{-1} B
        op_ground_up = invAup @ bop[..., :self.nup, :self.nup]
        op_ground_down = invAdown @ bop[..., self.nup:, :self.ndown]

        if op_squared:
            op_ground_up = op_ground_up @ op_ground_up
            op_ground_down = op_ground_down @ op_ground_down

        # ground state operator
        op_ground_up = btrace(op_ground_up)
        op_ground_down = btrace(op_ground_down)

        op_ground_up.unsqueeze_(-1)
        op_ground_down.unsqueeze_(-1)

        return op_ground_up, op_ground_down

    def operator_explicit(self, mo, bkin, op_squared=False):
        r"""Computes the value of any operator using the trace trick for a product
            of spin up/down determinant.

        .. math::
            -\\frac{1}{2} \Delta \Psi = -\\frac{1}{2}  D_{up} D_{down}
            ( \Delta_{up} D_{up} / D_{up} + \Delta_{down} D_{down}  / D_{down} )

        Args:
            mo (torch.tensor): matrix of MO vals(Nbatch, Nelec, Nmo)
            bkin (torch.tensor): kinetic operator (Nbatch, Nelec, Nmo)
            op_squared (bool, optional) return the trace of the square of the product if True

        Returns:
            torch.tensor: kinetic energy
        """

        # shortcut up/down matrices
        Aup, Adown = self.orb_proj.split_orbitals(mo)
        Bup, Bdown = self.orb_proj.split_orbitals(bkin)

        # check ifwe have 1 or multiple ops
        multiple_op = (Bup.ndim == 5)

        # inverse of MO matrices
        iAup = torch.inverse(Aup)
        iAdown = torch.inverse(Adown)

        # if we have multiple operators
        if multiple_op:
            iAup = iAup.unsqueeze(1)
            iAdown = iAdown.unsqueeze(1)

        # precompute product invA x B
        op_val_up = iAup @ Bup
        op_val_down = iAdown @ Bdown

        if op_squared:
            op_val_up = op_val_up @ op_val_up
            op_val_down = op_val_down @ op_val_down

        # kinetic terms
        op_val_up = btrace(op_val_up)
        op_val_down = btrace(op_val_down)

        # reshape
        if multiple_op:
            op_val_up = op_val_up.permute(1, 2, 0)
            op_val_down = op_val_down.permute(1, 2, 0)
        else:
            op_val_up = op_val_up.transpose(0, 1)
            op_val_down = op_val_down.transpose(0, 1)

        return (op_val_up, op_val_down)

    def operator_single_double(self, mo, bop, op_squared=False):
        """Computes the value of any operator on gs + single + double

        Args:
            mo (torch.tensor): matrix of molecular orbitals
            bkin (torch.tensor): matrix of kinetic operator
            op_squared (bool, optional) return the trace of the square of the product if True

        Returns:
            torch.tensor: kinetic energy values
        """

        op_up, op_down = self.operator_unique_single_double(
            mo, bop, op_squared)

        return (op_up[..., self.index_unique_excitation[0]],
                op_down[..., self.index_unique_excitation[1]])

    def operator_unique_single_double(self, mo, bop, op_squared):
        """Compute the operator value of the unique single/double conformation

        Args:
            mo ([type]): [description]
            bkin ([type]): [description]
            op_squared (bool) return the trace of the square of the product
        """

        nbatch = mo.shape[0]

        if not hasattr(self.exc_mask, 'index_unique_single_up'):
            self.exc_mask.get_index_unique_single()

        if not hasattr(self.exc_mask, 'index_unique_double_up'):
            self.exc_mask.get_index_unique_double()

        do_single = len(self.exc_mask.index_unique_single_up) != 0
        do_double = len(self.exc_mask.index_unique_double_up) != 0

        # occupied orbital matrix + det and inv on spin up
        Aocc_up = mo[:, :self.nup, :self.nup]

        # occupied orbital matrix + det and inv on spin down
        Aocc_down = mo[:, self.nup:, :self.ndown]

        # inverse of the
        invAup = torch.inverse(Aocc_up)
        invAdown = torch.inverse(Aocc_down)

        # precompute invA @ B
        invAB_up = invAup @ bop[..., :self.nup, :self.nup]
        invAB_down = invAdown @ bop[..., self.nup:, :self.ndown]

        # ground state operator
        if op_squared:
            op_ground_up = btrace(invAB_up@invAB_up)
            op_ground_down = btrace(invAB_down@invAB_down)
        else:
            op_ground_up = btrace(invAB_up)
            op_ground_down = btrace(invAB_down)

        op_ground_up.unsqueeze_(-1)
        op_ground_down.unsqueeze_(-1)

        # store the kin terms we need
        op_out_up = op_ground_up.clone()
        op_out_down = op_ground_down.clone()

        # virtual orbital matrices spin up/down
        Avirt_up = mo[:, :self.nup, self.nup:self.index_max_orb_up]
        Avirt_down = mo[:, self.nup:,
                        self.ndown: self.index_max_orb_down]

        # compute the products of invA and Btilde
        mat_exc_up = (invAup @ Avirt_up)
        mat_exc_down = (invAdown @ Avirt_down)

        bop_up = bop[..., :self.nup, :self.index_max_orb_up]
        bop_occ_up = bop[..., :self.nup, :self.nup]
        bop_virt_up = bop[..., :self.nup,
                          self.nup:self.index_max_orb_up]

        bop_down = bop[:, self.nup:, :self.index_max_orb_down]
        bop_occ_down = bop[..., self.nup:, :self.ndown]
        bop_virt_down = bop[..., self.nup:,
                            self.ndown:self.index_max_orb_down]

        Mup = invAup @ bop_virt_up - invAup @ bop_occ_up @ invAup @ Avirt_up
        Mdown = invAdown @ bop_virt_down - \
            invAdown @ bop_occ_down @ invAdown @ Avirt_down

        # if we only want the normal value of the op and not its squared
        if not op_squared:

            # reshape the M matrices
            Mup = Mup.view(*Mup.shape[:-2], -1)
            Mdown = Mdown.view(*Mdown.shape[:-2], -1)

            if do_single:

                # spin up
                op_sin_up = self.op_single(op_ground_up, mat_exc_up, Mup,
                                           self.exc_mask.index_unique_single_up, nbatch)

                # spin down
                op_sin_down = self.op_single(op_ground_down, mat_exc_down, Mdown,
                                             self.exc_mask.index_unique_single_down, nbatch)

                # store the terms we need
                op_out_up = torch.cat((op_out_up, op_sin_up), dim=-1)
                op_out_down = torch.cat(
                    (op_out_down, op_sin_down), dim=-1)

            if do_double:

                # spin up
                op_dbl_up = self.op_multiexcitation(op_ground_up, mat_exc_up, Mup,
                                                    self.exc_mask.index_unique_double_up,
                                                    2, nbatch)

                # spin down
                op_dbl_down = self.op_multiexcitation(op_ground_down, mat_exc_down, Mdown,
                                                      self.exc_mask.index_unique_double_down,
                                                      2, nbatch)

                # store the terms we need
                op_out_up = torch.cat((op_out_up, op_dbl_up), dim=-1)
                op_out_down = torch.cat(
                    (op_out_down, op_dbl_down), dim=-1)

            return op_out_up, op_out_down

        # if we watn the squre of the operatore
        # typically trace(ABAB)
        else:

            # compute A^-1 B M
            Yup = invAB_up @ Mup
            Ydown = invAB_down @ Mdown

            # reshape the M matrices
            Mup = Mup.view(*Mup.shape[:-2], -1)
            Mdown = Mdown.view(*Mdown.shape[:-2], -1)

            # reshape the Y matrices
            Yup = Yup.view(*Yup.shape[:-2], -1)
            Ydown = Ydown.view(*Ydown.shape[:-2], -1)

            if do_single:

                # spin up
                op_sin_up = self.op_squared_single(op_ground_up, mat_exc_up,
                                                   Mup, Yup,
                                                   self.exc_mask.index_unique_single_up,
                                                   nbatch)

                # spin down
                op_sin_down = self.op_squared_single(op_ground_down, mat_exc_down,
                                                     Mdown, Ydown,
                                                     self.exc_mask.index_unique_single_down,
                                                     nbatch)

                # store the terms we need
                op_out_up = torch.cat((op_out_up, op_sin_up), dim=-1)
                op_out_down = torch.cat(
                    (op_out_down, op_sin_down), dim=-1)

            if do_double:

                # spin up values
                op_dbl_up = self.op_squared_multiexcitation(op_ground_up, mat_exc_up,
                                                            Mup, Yup,
                                                            self.exc_mask.index_unique_double_down,
                                                            2, nbatch)

                # spin down values
                op_dbl_down = self.op_squared_multiexcitation(op_ground_down, mat_exc_down,
                                                              Mdown, Ydown,
                                                              self.exc_mask.index_unique_double_down,
                                                              2, nbatch)

                # store the terms we need
                op_out_up = torch.cat((op_out_up, op_dbl_up), dim=-1)
                op_out_down = torch.cat(
                    (op_out_down, op_dbl_down), dim=-1)

            return op_out_up, op_out_down

    @staticmethod
    def op_single(baseterm, mat_exc, M, index, nbatch):
        r"""Computes the operator values for single excitation

        .. math::
            Tr( \bar{A}^{-1} \bar{B}) = Tr(A^{-1} B) + Tr( T M )
            T = P ( A^{-1} \bar{A})^{-1} P
            M = A^{-1}\bar{B} - A^{-1}BA^{-1}\bar{A}

        Args:
            baseterm (torch.tensor): trace(A B)
            mat_exc (torch.tensor): invA @ Abar
            M (torch.tensor): invA Bbar - inv A B inv A Abar
            index(List): list of index of the excitations
            nbatch : batch size
        """

        # compute the values of T
        T = (1. / mat_exc.view(nbatch, -1)[:, index])

        # computes trace(T M)
        op_vals = T * M[..., index]

        # add the base terms
        op_vals += baseterm

        return op_vals

    @staticmethod
    def op_multiexcitation(baseterm, mat_exc, M, index, size, nbatch):
        r"""Computes the operator values for single excitation

        .. math::
            Tr( \bar{A}^{-1} \bar{B}) = Tr(A^{-1} B) + Tr( T M )
            T = P ( A^{-1} \bar{A})^{-1} P
            M = A^{-1}\bar{B} - A^{-1}BA^{-1}\bar{A}

        Args:
            baseterm (torch.tensor): trace(A B)
            mat_exc (torch.tensor): invA @ Abar
            M (torch.tensor): invA Bbar - inv A B inv A Abar
            index(List): list of index of the excitations
            size(int) : number of excitation
            nbatch : batch size
        """

        # get the values of the excitation matrix invA Abar
        T = mat_exc.view(nbatch, -1)[:, index]

        # get the shapes of the size x size matrices
        _ext_shape = (*T.shape[:-1], -1, size, size)
        _m_shape = (*M.shape[:-1], -1, size, size)

        # computes the inverse of invA Abar
        T = torch.inverse(T.view(_ext_shape))

        # computes T @ M (after reshaping M as size x size matrices)
        op_vals = T @ (M[..., index]).view(_m_shape)

        # compute the trace
        op_vals = btrace(op_vals)

        # add the base term
        op_vals += baseterm

        return op_vals

    @staticmethod
    def op_squared_single(baseterm, mat_exc, M, Y, index, nbatch):
        r"""Computes the operator squared for single excitation

        .. math::
            Tr( (\bar{A}^{-1} \bar{B})^2) = Tr((A^{-1} B)^2) + Tr( (T M)^2 ) + 2 Tr(T Y)
            T = P ( A^{-1} \bar{A})^{-1} P -> mat_exc in the code
            M = A^{-1}\bar{B} - A^{-1}BA^{-1}\bar{A}
            Y = A^{-1} B M

        Args:
            baseterm (torch.tensor): trace(A B A B)
            mat_exc (torch.tensor): invA @ Abar
            M (torch.tensor): invA Bbar - inv A B inv A Abar
            Y (torch.tensor): invA B M
            index(List): list of index of the excitations
            nbatch : batch size
        """

        # get the values of the inverse excitation matrix
        T = 1. / (mat_exc.view(nbatch, -1)[:, index])

        # compute  trace(( T M )^2)
        tmp = (T * M[..., index])
        op_vals = tmp*tmp

        # trace(T Y)
        tmp = (T * Y[..., index])
        op_vals += 2 * tmp

        # add the base term
        op_vals += baseterm

        return op_vals

    @staticmethod
    def op_squared_multiexcitation(baseterm, mat_exc, M, Y, index, size, nbatch):
        r"""Computes the operator squared for multiple excitation

        .. math::
            Tr( (\bar{A}^{-1} \bar{B})^2) = Tr((A^{-1} B)^2) + Tr( (T M)^2 ) + 2 Tr(T Y)
            T = P ( A^{-1} \bar{A})^{-1} P -> mat_exc in the code
            M = A^{-1}\bar{B} - A^{-1}BA^{-1}\bar{A}
            Y = A^{-1} B M

        Args:
            baseterm (torch.tensor): trace(A B A B)
            mat_exc (torch.tensor): invA @ Abar
            M (torch.tensor): invA Bbar - inv A B inv A Abar
            Y (torch.tensor): invA B M
            index(List): list of index of the excitations
            nbatch : batch size
            size(int): number of excitation
        """

        # get the values of the excitation matrix invA Abar
        T = mat_exc.view(nbatch, -1)[:, index]

        # get the shape as a series of size x size matrices
        _ext_shape = (*T.shape[:-1], -1, size, size)
        _m_shape = (*M.shape[:-1], -1, size, size)
        _y_shape = (*Y.shape[:-1], -1, size, size)

        # reshape T and take the inverse of the matrices
        T = torch.inverse(T.view(_ext_shape))

        # compute  trace(( T M )^2)
        tmp = T @ (M[..., index]).view(_m_shape)

        # take the trace of that and add to base value
        tmp = btrace(tmp @ tmp)
        op_vals = tmp

        # compute trace( T Y )
        tmp = T @ (Y[..., index]).view(_y_shape)
        tmp = btrace(tmp)
        op_vals += 2*tmp

        # add the base term
        op_vals += baseterm

        return op_vals



from types import SimpleNamespace
import itertools
import numpy as np
from pyscf import gto, scf, dft
import shutil
from .calculator_base import CalculatorBase


class CalculatorPySCF(CalculatorBase):

    def __init__(self, atoms, atom_coords, basis, scf, units, molname, savefile):

        CalculatorBase.__init__(
            self, atoms, atom_coords, basis, scf, units, molname, 'pyscf', savefile)

    def run(self):
        """Run the scf calculation using PySCF."""

        # refresh the atom positions if necessary
        atom_str = self.get_atoms_str()

        # pyscf calculation
        mol = gto.M(
            atom=atom_str,
            basis=self.basis_name,
            unit=self.units,
            cart=False)

        if self.scf.lower() == 'hf':
            pyscf_data = scf.RHF(mol).run()

        elif self.scf.lower() == 'dft':
            pyscf_data = dft.RKS(mol)
            pyscf_data.xc = 'lda, vwn'
            pyscf_data = pyscf_data.newton()
            pyscf_data.kernel()

        if self.savefile:
            save_file_name = self.molname + '_pyscf.chkfile'
            shutil.copyfile(pyscf_data.chkfile, save_file_name)
            self.savefile = save_file_name

        basis = self.get_basis_data(mol, pyscf_data)
        return basis

    def get_basis_data(self, mol, rhf):
        """Save the data to HDF5

        Arguments:
            mol {pyscf.gto.M} -- psycf Molecule
            rhf {pyscf.scf} -- scf object
        """

        kx = {0: [0], 1: [1, 0, 0], 2: [2, 1, 1, 0, 0, 0]}
        ky = {0: [0], 1: [0, 1, 0], 2: [0, 1, 0, 2, 1, 0]}
        kz = {0: [0], 1: [0, 0, 1], 2: [0, 0, 1, 0, 1, 2]}

        basis = SimpleNamespace()
        basis.TotalEnergy = rhf.e_tot
        basis.radial_type = 'gto_pure'
        basis.harmonics_type = 'cart'

        # number of AO / MO
        # can be different if d or f orbs are present
        # due to cart 2 sph harmonics
        basis.nao = len(mol.cart_labels())
        basis.nmo = rhf.mo_energy.shape[0]

        # nshells is the number of bas per atoms
        nshells = [0] * mol.natm

        # init bas properties
        bas_coeff, bas_exp = [], []
        index_ctr = []
        bas_n, bas_l = [], []
        bas_kx, bas_ky, bas_kz = [], [], []
        bas_n = []
        bas_n_ori = self.get_bas_n(mol)

        iao = 0
        for ibas in range(mol.nbas):

            # number of contracted gaussian in that bas
            # nctr = mol.bas_nctr(ibas)
            nctr = mol.bas_nprim(ibas)

            # number of ao from that bas
            mult = mol.bas_len_cart(ibas)

            # quantum numbers
            n = bas_n_ori[ibas]
            lval = mol.bas_angular(ibas)

            # get qn per bas
            bas_n += [n] * nctr * mult
            bas_l += [lval] * nctr * mult

            # coeffs/exp
            bas_coeff += mol.bas_ctr_coeff(
                ibas).flatten().tolist() * mult
            bas_exp += mol.bas_exp(ibas).flatten().tolist() * mult

            # number of shell per atoms
            nshells[mol.bas_atom(ibas)] += nctr * mult

            for m in range(mult):
                index_ctr += [iao] * nctr
                iao += 1

            for k in kx[lval]:
                bas_kx += [k] * nctr

            for k in ky[lval]:
                bas_ky += [k] * nctr

            for k in kz[lval]:
                bas_kz += [k] * nctr

        bas_norm = []
        for expnt, lval in zip(bas_exp, bas_l):
            bas_norm.append(mol.gto_norm(lval, expnt))

        basis.nshells = nshells
        basis.index_ctr = index_ctr
        intervals = np.concatenate(([0], np.cumsum(nshells)))

        basis.nao_per_atom = []
        for i in range(len(intervals)-1):
            s, e = intervals[i], intervals[i+1]
            nao = len(np.unique(basis.index_ctr[s:e]))
            basis.nao_per_atom.append(nao)

        # determine the number of contraction per
        # atomic orbital
        basis.nctr_per_ao = np.array(
            [len(list(y)) for _, y in itertools.groupby(index_ctr)])

        basis.bas_coeffs = np.array(bas_coeff)
        basis.bas_exp = np.array(bas_exp)
        basis.bas_norm = np.array(bas_norm)

        basis.bas_n = bas_n
        basis.bas_l = bas_l

        # the cartesian gto are all :
        #   x^a y^b z^c exp(-zeta r)
        # i.e. there is no r dependency
        basis.bas_kr = np.zeros_like(basis.bas_exp)

        basis.bas_kx = np.array(bas_kx)
        basis.bas_ky = np.array(bas_ky)
        basis.bas_kz = np.array(bas_kz)

        # molecular orbitals
        basis.mos = mol.cart2sph_coeff() @ rhf.mo_coeff
        basis.mos = self.normalize_columns(basis.mos)

        # atom coords
        basis.atom_coords_internal = self.atom_coords

        return basis

    def get_atoms_str(self):
        """Refresh the atom string (use after atom move). """
        atoms_str = ''
        natom = len(self.atoms)

        for iA in range(natom):
            atoms_str += self.atoms[iA] + ' '
            atoms_str += ' '.join(str(xi)
                                  for xi in self.atom_coords[iA])
            atoms_str += ';'
        return atoms_str

    @staticmethod
    def get_bas_n(mol):

        label2int = {'s': 1, 'p': 2, 'd': 3}
        labels = [l[:3] for l in mol.cart_labels(fmt=False)]
        unique_labels = []
        for l in labels:
            if l not in unique_labels:
                unique_labels.append(l)
        nlabel = [l[2][1] for l in unique_labels]
        n = [label2int[nl] for nl in nlabel]
        return n



import torch
import numpy as np


def atomic_orbital_norm(basis):
    """Computes the norm of the atomic orbitals

    Args:
        basis (Namespace): basis object of the Molecule instance

    Returns:
        torch.tensor: Norm of the atomic orbitals

    Examples::
        >>> mol = Molecule('h2.xyz', basis='dzp', calculator='adf')
        >>> norm = atomic_orbital_norm(mol.basis)
    """

    # spherical
    if basis.harmonics_type == 'sph':

        if basis.radial_type.startswith('sto'):
            return norm_slater_spherical(basis.bas_n, basis.bas_exp)

        elif basis.radial_type.startswith('gto'):
            return norm_gaussian_spherical(basis.bas_n, basis.bas_exp)

        else:
            raise ValueError('%s is not a valid radial_type')

    # cartesian
    elif basis.harmonics_type == 'cart':

        if basis.radial_type.startswith('sto'):
            return norm_slater_cartesian(
                basis.bas_kx,
                basis.bas_ky,
                basis.bas_kz,
                basis.bas_kr,
                basis.bas_exp)

        elif basis.radial_type.startswith('gto'):
            return norm_gaussian_cartesian(
                basis.bas_kx, basis.bas_ky, basis.bas_kz, basis.bas_exp)

        else:
            raise ValueError('%s is not a valid radial_type')


def norm_slater_spherical(bas_n, bas_exp):
    """Normalization of STOs with Sphecrical Harmonics. \n
     * www.theochem.ru.nl/~pwormer/Knowino/knowino.org/wiki/Slater_orbital \n
     * C Filippi,  JCP 105, 213 1996 \n
     * Monte Carlo Methods in Ab Inition Quantum Chemistry, B.L. Hammond

    Args:
        bas_n (torch.tensor): prinicpal quantum number
        bas_exp (torch.tensor): slater exponents

    Returns:
        torch.tensor: normalization factor
    """
    nfact = torch.as_tensor([np.math.factorial(2 * n)
                             for n in bas_n], dtype=torch.get_default_dtype())
    return (2 * bas_exp)**bas_n * torch.sqrt(2 * bas_exp / nfact)


def norm_gaussian_spherical(bas_n, bas_exp):
    """Normlization of GTOs with spherical harmonics. \n
     * Computational Quantum Chemistry: An interactive Intrduction to basis set theory \n
        eq : 1.14 page 23.

    Args:
        bas_n (torch.tensor): prinicpal quantum number
        bas_exp (torch.tensor): slater exponents

    Returns:
        torch.tensor: normalization factor
    """

    from scipy.special import factorial2 as f2

    bas_n = bas_n + 1.
    exp1 = 0.25 * (2. * bas_n + 1.)

    A = bas_exp**exp1
    B = 2**(2. * bas_n + 3. / 2)
    C = torch.as_tensor(f2(2 * bas_n.int() - 1) * np.pi **
                        0.5).type(torch.get_default_dtype())

    return torch.sqrt(B / C) * A


def norm_slater_cartesian(a, b, c, n, exp):
    """Normaliation of STos with cartesian harmonics. \n
     * Monte Carlo Methods in Ab Initio Quantum Chemistry page 279

    Args:
        a (torch.tensor): exponent of x
        b (torch.tensor): exponent of y
        c (torch.tensor): exponent of z
        n (torch.tensor): exponent of r
        exp (torch.tensor): Sater exponent

    Returns:
        torch.tensor: normalization factor
    """
    from scipy.special import factorial2 as f2

    lvals = a + b + c + n + 1.

    lfact = torch.as_tensor([np.math.factorial(2 * i)
                             for i in lvals]).type(torch.get_default_dtype())

    prefact = 4 * np.pi * lfact / ((2 * exp)**(2 * lvals + 1))

    num = torch.as_tensor(f2(2 * a.astype('int') - 1) *
                          f2(2 * b.astype('int') - 1) *
                          f2(2 * c.astype('int') - 1)
                          ).type(torch.get_default_dtype())

    denom = torch.as_tensor(
        f2((2 * a + 2 * b + 2 * c + 1).astype('int')
           )).type(torch.get_default_dtype())

    return torch.sqrt(1. / (prefact * num / denom))


def norm_gaussian_cartesian(a, b, c, exp):
    """Normaliation of GTOs with cartesian harmonics. \n
     * Monte Carlo Methods in Ab Initio Quantum Chemistry page 279

    Args:
        a (torch.tensor): exponent of x
        b (torch.tensor): exponent of y
        c (torch.tensor): exponent of z
        exp (torch.tensor): Sater exponent

    Returns:
        torch.tensor: normalization factor
    """

    from scipy.special import factorial2 as f2

    pref = torch.as_tensor((2 * exp / np.pi)**(0.75))
    am1 = (2 * a - 1).astype('int')
    x = (4 * exp)**(a / 2) / torch.sqrt(torch.as_tensor(f2(am1)))

    bm1 = (2 * b - 1).astype('int')
    y = (4 * exp)**(b / 2) / torch.sqrt(torch.as_tensor(f2(bm1)))

    cm1 = (2 * c - 1).astype('int')
    z = (4 * exp)**(c / 2) / torch.sqrt(torch.as_tensor(f2(cm1)))

    return (pref * x * y * z).type(torch.get_default_dtype())



import torch
from torch import nn
from .backflow_kernel_base import BackFlowKernelBase


class BackFlowKernelInverse(BackFlowKernelBase):

    def __init__(self, mol, cuda=False):
        """Compute the back flow kernel, i.e. the function
        f(rij) where rij is the distance between electron i and j
        This kernel is used in the backflow transformation
        .. math:
            q_i = r_i + \\sum_{j\\neq i} f(r_{ij}) (r_i-r_j)

        with here :

        .. math:
            f(r_{ij) = \\frac{w}{r_{ij}
        """
        super().__init__(mol, cuda)
        self.weight = nn.Parameter(
            torch.as_tensor([1E-3]))  # .to(self.device)

    def _backflow_kernel(self, ree):
        """Computes the backflow kernel:

        .. math:
            \\eta(r_{ij}) = \\frac{w}{r_{ij}}

        Args:
            r (torch.tensor): e-e distance Nbatch x Nelec x Nelec

        Returns:
            torch.tensor : f(r) Nbatch x Nelec x Nelec
        """

        eye = torch.eye(self.nelec, self.nelec).to(self.device)
        mask = torch.ones_like(ree) - eye
        return self.weight * mask * (1./(ree+eye) - eye)

    def _backflow_kernel_derivative(self, ree):
        """Computes the derivative of the kernel function
            w.r.t r_{ij}
        .. math::
            \\frac{d}{dr_{ij} \\eta(r_{ij}) = -w r_{ij}^{-2}

        Args:
            ree (torch.tensor): e-e distance Nbatch x Nelec x Nelec

        Returns:
            torch.tensor : f'(r) Nbatch x Nelec x Nelec
        """

        eye = torch.eye(self.nelec, self.nelec).to(self.device)
        invree = (1./(ree+eye) - eye)
        return - self.weight * invree * invree

    def _backflow_kernel_second_derivative(self, ree):
        """Computes the derivative of the kernel function
            w.r.t r_{ij}
        .. math::
            \\frac{d^2}{dr_{ij}^2} \\eta(r_{ij}) = 2 w r_{ij}^{-3}

        Args:
            ree (torch.tensor): e-e distance Nbatch x Nelec x Nelec

        Returns:
            torch.tensor : f''(r) Nbatch x Nelec x Nelec
        """

        eye = torch.eye(self.nelec, self.nelec).to(self.device)
        invree = (1./(ree+eye) - eye)
        return 2 * self.weight * invree * invree * invree



import unittest

import numpy as np
import torch
import torch.optim as optim

from qmctorch.sampler import Metropolis
from qmctorch.solver import Solver
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterOrbitalDependentJastrow
from qmctorch.utils import set_torch_double_precision

from ..path_utils import PATH_TEST


class TestLiHCorrelated(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(0)
        np.random.seed(0)
        set_torch_double_precision()

        # molecule
        path_hdf5 = (
            PATH_TEST / 'hdf5/LiH_adf_dz.hdf5').absolute().as_posix()
        self.mol = Molecule(load=path_hdf5)

        # wave function
        self.wf = SlaterOrbitalDependentJastrow(self.mol,
                                                kinetic='jacobi',
                                                configs='cas(2,2)',
                                                include_all_mo=True)

        # fc weights
        self.wf.fc.weight.data = torch.rand(self.wf.fc.weight.shape)

        # jastrow weights
        for ker in self.wf.jastrow.jastrow_kernel.jastrow_functions:
            ker.weight.data = torch.rand(1)

        # sampler
        self.sampler = Metropolis(
            nwalkers=500,
            nstep=200,
            step_size=0.05,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        # optimizer
        self.opt = optim.Adam(self.wf.parameters(), lr=0.01)

        # solver
        self.solver = Solver(wf=self.wf,
                                          sampler=self.sampler,
                                          optimizer=self.opt)

        # artificial pos
        self.nbatch = 10
        self.pos = torch.as_tensor(np.random.rand(
            self.nbatch, self.wf.nelec*3))
        self.pos.requires_grad = True

    def test_0_wavefunction(self):

        eauto = self.wf.kinetic_energy_autograd(self.pos)
        ejac = self.wf.kinetic_energy_jacobi(self.pos)

        assert torch.allclose(
            eauto.data, ejac.data, rtol=1E-4, atol=1E-4)

    def test1_single_point(self):

        # sample and compute observables
        obs = self.solver.single_point()
        _, _ = obs.energy, obs.variance

    # def test2_wf_opt_grad_auto(self):
    #     self.solver.sampler = self.sampler

    #     self.solver.configure(track=['local_energy'],
    #                           loss='energy', grad='auto')
    #     obs = self.solver.run(5)

    def test3_wf_opt_grad_manual(self):
        self.solver.sampler = self.sampler

        self.solver.configure(track=['local_energy'],
                              loss='energy', grad='manual')
        obs = self.solver.run(5)


if __name__ == "__main__":
    unittest.main()
    # t = TestLiHCorrelated()
    # t.setUp()
    # t.test_0_wavefunction()
    # t.test1_single_point()
    # t.test2_wf_opt_grad_auto()
    # t.test3_wf_opt_grad_manual()



from .fully_connected_jastrow_kernel import FullyConnectedJastrowKernel
from .jastrow_kernel_electron_nuclei_base import JastrowKernelElectronNucleiBase
from .pade_jastrow_kernel import PadeJastrowKernel



from qmctorch.scf import Molecule
import numpy as np
import unittest


class TestMolecule(unittest.TestCase):

    def test1_create(self):

        # molecule
        mol = Molecule(
            atom='H 0. 0. 0.; H 0. 0. 1.',
            unit='bohr',
            scf='hf',
            calculator='pyscf',
            basis='sto-3g',
            redo_scf=True)

        mol.print_total_energy()

    def test2_load(self):
        mol = Molecule(load='H2_pyscf_sto-3g.hdf5')

    def test3_domain(self):
        mol = Molecule(load='H2_pyscf_sto-3g.hdf5')

        domain_center = mol.domain('center')
        assert (domain_center['center'] ==
                np.array([0., 0., 0.5])).all()

        domain_uniform = mol.domain('uniform')
        assert domain_uniform == {
            'method': 'uniform', 'min': -0.5, 'max': 1.5}

        domain_normal = mol.domain('normal')
        assert np.all(domain_normal['mean']
                      == np.array([0., 0., 0.5]))

        domain_atomic = mol.domain('atomic')
        assert np.all(domain_atomic['atom_coords'] == np.array([[0., 0., 0.],
                                                                [0., 0., 1.]]))

    def test4_create_dft(self):

        # molecule
        mol = Molecule(
            atom='H 0. 0. 0.; H 0. 0. 1.',
            unit='bohr',
            calculator='pyscf',
            scf='dft',
            basis='sto-3g',
            redo_scf=True)

        mol.print_total_energy()


if __name__ == "__main__":
    unittest.main()



import torch
from .jastrow_kernel_electron_electron_nuclei_base import JastrowKernelElectronElectronNucleiBase


class FullyConnectedJastrowKernel(JastrowKernelElectronElectronNucleiBase):

    def __init__(self, nup, ndown, atomic_pos, cuda):
        """Defines a fully connected jastrow factors."""

        super().__init__(nup, ndown, atomic_pos, cuda)

        self.fc1 = torch.nn.Linear(3, 9, bias=True)
        self.fc2 = torch.nn.Linear(9, 3, bias=True)
        self.fc3 = torch.nn.Linear(3, 1, bias=True)

        torch.nn.init.uniform_(self.fc1.weight)
        torch.nn.init.uniform_(self.fc2.weight)
        torch.nn.init.uniform_(self.fc2.weight)

        self.fc1.weight.data *= 1E-3
        self.fc2.weight.data *= 1E-3
        self.fc3.weight.data *= 1E-3

        self.nl_func = torch.nn.Sigmoid()

    def forward(self, x):
        """Compute the values of the individual f_ij=f(r_ij)

        Args:
            x (torch.tensor): e-e distance Nbatch, Nele_pairs

        Returns:
            torch.tensor: values of the f_ij
        """

        # reshape the input so that all elements
        # are considered independently of each other
        out_shape = list(x.shape)[:-1] + [1]
        x = x.reshape(-1, 3)

        x = self.fc1(x)
        x = 2 * self.nl_func(x)
        x = self.fc2(x)
        x = 2 * self.nl_func(x)
        x = self.fc3(x)
        x = 2 * self.nl_func(x)

        return x.reshape(*out_shape)



import torch
from torch import nn
from torch.autograd import grad
from torch.autograd.variable import Variable


class JastrowKernelElectronElectronNucleiBase(nn.Module):

    def __init__(self, nup, ndown, atomic_pos, cuda, **kwargs):
        r"""Base Class for the elec-elec-nuc jastrow kernel


        Args:
            nup (int): number of spin up electons
            ndow (int): number of spin down electons
            atoms (torch.tensor): atomic positions of the atoms
            cuda (bool, optional): Turns GPU ON/OFF. Defaults to False.
        """

        super().__init__()
        self.nup, self.ndown = nup, ndown
        self.cuda = cuda

        self.nelec = nup + ndown
        self.atoms = atomic_pos
        self.natoms = atomic_pos.shape[0]
        self.ndim = 3

        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')
        self.requires_autograd = True

    def forward(self, x):
        """Compute the values of the kernel

        Args:
            x (torch.tensor): e-e and e-n distances distance (Nbatch, Natom, Nelec_pairs, 3)
                              the last dimension holds the values [R_{iA}, R_{jA}, r_{ij}]
                              in that order.

        Returns:
            torch.tensor: values of the kernel (Nbatch, Natom, Nelec_pairs, 1)

        """
        raise NotImplementedError()

    def compute_derivative(self, r, dr):
        """Get the elements of the derivative of the jastrow kernels."""

        kernel = self.forward(r)
        ker_grad = self._grads(kernel, r)

        # return the ker * dr
        out = ker_grad.unsqueeze(1) * dr

        # sum over the atoms
        return out

    def compute_second_derivative(self, r, dr, d2r):
        """Get the elements of the pure 2nd derivative of the jastrow kernels.
        """

        dr2 = dr * dr

        kernel = self.forward(r)
        ker_hess, ker_grad = self._hess(kernel, r, self.device)

        jhess = ker_hess.unsqueeze(1) * \
            dr2 + ker_grad.unsqueeze(1) * d2r

        return jhess

    @staticmethod
    def _grads(val, pos):
        """Get the gradients of the jastrow values
        of a given orbital terms

        Args:
            pos ([type]): [description]

        Returns:
            [type]: [description]
        """
        return grad(val, pos, grad_outputs=torch.ones_like(val))[0]

    @staticmethod
    def _hess(val, pos, device):
        """get the hessian of the jastrow values.

        Args:
            pos ([type]): [description]
        """

        gval = grad(val, pos,
                    grad_outputs=torch.ones_like(val),
                    create_graph=True)[0]

        grad_out = Variable(torch.ones(
            *gval.shape[:-1])).to(device)
        hval = torch.zeros_like(gval).to(device)

        for idim in range(gval.shape[-1]):

            tmp = grad(gval[..., idim], pos,
                       grad_outputs=grad_out,
                       only_inputs=True,
                       create_graph=True)[0]
            hval[..., idim] = tmp[..., idim]

        return hval, gval



import torch
from torch import nn
from ..distance.electron_nuclei_distance import ElectronNucleiDistance


class JastrowFactorElectronNuclei(nn.Module):

    def __init__(self, nup, ndown, atomic_pos,
                 jastrow_kernel,
                 kernel_kwargs={},
                 cuda=False):
        r"""Base class for two el-nuc jastrow of the form:

        .. math::
            J = \prod_{a,i} \exp(A(r_{ai}))

        Args:
            nup (int): number of spin up electons
            ndow (int): number of spin down electons
            atomic_pos (tensor): positions of the atoms
            cuda (bool, optional): Turns GPU ON/OFF. Defaults to False.
        """

        super().__init__()

        self.nup = nup
        self.ndown = ndown
        self.nelec = nup + ndown

        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')

        self.atoms = atomic_pos.to(self.device)
        self.natoms = atomic_pos.shape[0]
        self.ndim = 3

        # kernel function
        self.jastrow_kernel = jastrow_kernel(nup, ndown,
                                             atomic_pos, cuda,
                                             **kernel_kwargs)

        # requires autograd to compute derivatives
        self.requires_autograd = self.jastrow_kernel.requires_autograd

        # elec-nuc distances
        self.edist = ElectronNucleiDistance(
            self.nelec, self.atoms, self.ndim)

    def forward(self, pos, derivative=0, sum_grad=True):
        """Compute the Jastrow factors.

        Args:
            pos (torch.tensor): Positions of the electrons
                                  Size : Nbatch, Nelec x Ndim
            derivative (int, optional): order of the derivative (0,1,2,).
                            Defaults to 0.
            sum_grad (bool, optional): Return the sum_grad (i.e. the sum of
                                       the derivatives) or the individual
                                       terms. Defaults to True.
                                       False only for derivative=1

        Returns:
            torch.tensor: value of the jastrow parameter for all confs
                          derivative = 0  (Nmo) x Nbatch x 1
                          derivative = 1  (Nmo) x Nbatch x Nelec (for sum_grad = True)
                          derivative = 1  (Nmo) x Nbatch x Ndim x Nelec (for sum_grad = False)
                          derivative = 2  (Nmo) x Nbatch x Nelec
        """

        size = pos.shape
        assert size[1] == self.nelec * self.ndim

        r = self.edist(pos)
        kern_vals = self.jastrow_kernel(r)
        jast = torch.exp(kern_vals.sum([-1, -2])).unsqueeze(-1)

        if derivative == 0:
            return jast

        elif derivative == 1:
            dr = self.edist(pos, derivative=1)
            return self.jastrow_factor_derivative(r, dr, jast, sum_grad)

        elif derivative == 2:

            dr = self.edist(pos, derivative=1)
            d2r = self.edist(pos, derivative=2)

            return self.jastrow_factor_second_derivative(r, dr, d2r, jast)

        elif derivative == [0, 1, 2]:

            dr = self.edist(pos, derivative=1)
            d2r = self.edist(pos, derivative=2)

            return(jast,
                   self.jastrow_factor_derivative(
                       r, dr, jast, sum_grad),
                   self.jastrow_factor_second_derivative(r, dr, d2r, jast))

    def jastrow_factor_derivative(self, r, dr, jast, sum_grad):
        """Compute the value of the derivative of the Jastrow factor

        Args:
            r (torch.tensor): ee distance matrix Nbatch x Nelec x Nelec
            jast (torch.tensor): values of the jastrow elements
                                 Nbatch x Nelec x Natom

        Returns:
            torch.tensor: gradient of the jastrow factors
                          Nbatch x Ndim x Nelec
        """
        nbatch = r.shape[0]
        if sum_grad:

            djast = self.jastrow_kernel.compute_derivative(
                r, dr).sum((1, 3))
            return djast * jast
        else:

            djast = self.jastrow_kernel.compute_derivative(
                r, dr).sum(3)
            return djast * jast.unsqueeze(-1)

    def jastrow_factor_second_derivative(self, r, dr, d2r, jast):
        """Compute the value of the pure 2nd derivative of the Jastrow factor

        Args:
            r (torch.tensor): ee distance matrix Nbatch x Nelec x Nelec
            jast (torch.tensor): values of the ajstrow elements
                                 Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: diagonal hessian of the jastrow factors
                          Nbatch x Nelec x Ndim
        """
        nbatch = r.shape[0]

        # pure second derivative terms
        d2jast = self.jastrow_kernel.compute_second_derivative(
            r, dr, d2r).sum((1, 3))

        # mixed terms
        djast = self.jastrow_kernel.compute_derivative(r, dr)
        djast = ((djast.sum(3))**2).sum(1)

        # add partial derivative
        hess_jast = d2jast + djast

        return hess_jast * jast



import torch
from torch import nn

from .norm_orbital import atomic_orbital_norm
from .radial_functions import (radial_gaussian, radial_gaussian_pure,
                               radial_slater, radial_slater_pure)
from .spherical_harmonics import Harmonics


class AtomicOrbitals(nn.Module):

    def __init__(self, mol, cuda=False):
        """Computes the value of atomic orbitals

        Args:
            mol (Molecule): Molecule object
            cuda (bool, optional): Turn GPU ON/OFF Defaults to False.
        """

        super(AtomicOrbitals, self).__init__()
        dtype = torch.get_default_dtype()

        # wavefunction data
        self.nelec = mol.nelec
        self.norb = mol.basis.nao
        self.ndim = 3

        # make the atomic position optmizable
        self.atom_coords = nn.Parameter(torch.as_tensor(
            mol.basis.atom_coords_internal).type(dtype))
        self.atom_coords.requires_grad = True
        self.natoms = len(self.atom_coords)
        self.atomic_number = mol.atomic_number

        # define the BAS positions.
        self.nshells = torch.as_tensor(mol.basis.nshells)
        self.nao_per_atom = torch.as_tensor(mol.basis.nao_per_atom)
        self.bas_coords = self.atom_coords.repeat_interleave(
            self.nshells, dim=0)
        self.nbas = len(self.bas_coords)

        # index for the contractions
        self.index_ctr = torch.as_tensor(mol.basis.index_ctr)
        self.nctr_per_ao = torch.as_tensor(mol.basis.nctr_per_ao)
        self.contract = not len(torch.unique(
            self.index_ctr)) == len(self.index_ctr)

        # get the coeffs of the bas
        self.bas_coeffs = torch.as_tensor(
            mol.basis.bas_coeffs).type(dtype)

        # get the exponents of the bas
        self.bas_exp = nn.Parameter(
            torch.as_tensor(mol.basis.bas_exp).type(dtype))
        self.bas_exp.requires_grad = True

        # harmonics generator
        self.harmonics_type = mol.basis.harmonics_type
        if mol.basis.harmonics_type == 'sph':
            self.bas_n = torch.as_tensor(mol.basis.bas_n).type(dtype)
            self.harmonics = Harmonics(
                mol.basis.harmonics_type,
                bas_l=mol.basis.bas_l,
                bas_m=mol.basis.bas_m,
                cuda=cuda)

        elif mol.basis.harmonics_type == 'cart':
            self.bas_n = torch.as_tensor(mol.basis.bas_kr).type(dtype)
            self.harmonics = Harmonics(
                mol.basis.harmonics_type,
                bas_kx=mol.basis.bas_kx,
                bas_ky=mol.basis.bas_ky,
                bas_kz=mol.basis.bas_kz,
                cuda=cuda)

        # select the radial apart
        radial_dict = {'sto': radial_slater,
                       'gto': radial_gaussian,
                       'sto_pure': radial_slater_pure,
                       'gto_pure': radial_gaussian_pure}
        self.radial = radial_dict[mol.basis.radial_type]
        self.radial_type = mol.basis.radial_type

        # get the normalisation constants
        if hasattr(mol.basis, 'bas_norm') and False:
            self.norm_cst = torch.as_tensor(
                mol.basis.bas_norm).type(dtype)
        else:
            with torch.no_grad():
                self.norm_cst = atomic_orbital_norm(
                    mol.basis).type(dtype)

        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self._to_device()

    def __repr__(self):
        name = self.__class__.__name__
        return name + '(%s, %s, %d -> (%d,%d) )' % (self.radial_type, self.harmonics_type,
                                                    self.nelec*self.ndim, self.nelec,
                                                    self.norb)

    def _to_device(self):
        """Export the non parameter variable to the device."""

        self.device = torch.device('cuda')
        self.to(self.device)
        attrs = ['bas_n', 'bas_coeffs',
                 'nshells', 'norm_cst',
                 'index_ctr', 'nctr_per_ao',
                 'nao_per_atom']
        for at in attrs:
            self.__dict__[at] = self.__dict__[at].to(self.device)

    def forward(self, pos, derivative=[0], sum_grad=True, sum_hess=True, one_elec=False):
        """Computes the values of the atomic orbitals.

        .. math::
            \phi_i(r_j) = \sum_n c_n \\text{Rad}^{i}_n(r_j) \\text{Y}^{i}_n(r_j)

        where Rad is the radial part and Y the spherical harmonics part.
        It is also possible to compute the first and second derivatives

        .. math::
            \\nabla \phi_i(r_j) = \\frac{d}{dx_j} \phi_i(r_j) + \\frac{d}{dy_j} \phi_i(r_j) + \\frac{d}{dz_j} \phi_i(r_j) \n
            \\text{grad} \phi_i(r_j) = (\\frac{d}{dx_j} \phi_i(r_j), \\frac{d}{dy_j} \phi_i(r_j), \\frac{d}{dz_j} \phi_i(r_j)) \n
            \Delta \phi_i(r_j) = \\frac{d^2}{dx^2_j} \phi_i(r_j) + \\frac{d^2}{dy^2_j} \phi_i(r_j) + \\frac{d^2}{dz^2_j} \phi_i(r_j)

        Args:
            pos (torch.tensor): Positions of the electrons
                                  Size : Nbatch, Nelec x Ndim
            derivative (int, optional): order of the derivative (0,1,2,).
                                        Defaults to 0.
            sum_grad (bool, optional): Return the sum_grad (i.e. the sum of
                                       the derivatives) or the individual
                                       terms. Defaults to True.
                                       False only for derivative=1

            sum_hess (bool, optional): Return the sum_hess (i.e. the sum of
                                       2nd the derivatives) or the individual
                                       terms. Defaults to True.
                                       False only for derivative=1

            one_elec (bool, optional): if only one electron is in input

        Returns:
            torch.tensor: Value of the AO (or their derivatives) \n
                          size : Nbatch, Nelec, Norb (sum_grad = True) \n
                          size : Nbatch, Nelec, Norb, Ndim (sum_grad = False)

        Examples::
            >>> mol = Molecule('h2.xyz')
            >>> ao = AtomicOrbitals(mol)
            >>> pos = torch.rand(100,6)
            >>> aovals = ao(pos)
            >>> daovals = ao(pos,derivative=1)
        """

        if not isinstance(derivative, list):
            derivative = [derivative]

        if not sum_grad:
            assert(1 in derivative)

        if not sum_hess:
            assert(2 in derivative)

        if one_elec:
            nelec_save = self.nelec
            self.nelec = 1

        if derivative == [0]:
            ao = self._compute_ao_values(pos)

        elif derivative == [1]:
            ao = self._compute_first_derivative_ao_values(
                pos, sum_grad)

        elif derivative == [2]:
            ao = self._compute_second_derivative_ao_values(
                pos, sum_hess)

        elif derivative == [3]:
            ao = self._compute_mixed_second_derivative_ao_values(pos)

        elif derivative == [0, 1, 2]:
            ao = self._compute_all_ao_values(pos)

        else:
            raise ValueError(
                'derivative must be 0, 1, 2, 3 or [0, 1, 2, 3], got ', derivative)

        if one_elec:
            self.nelec = nelec_save

        return ao

    def _compute_ao_values(self, pos):
        """Compute the value of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, NelexNdim

        Returns:
            torch.tensor: atomic orbital values size (Nbatch, Nelec, Norb)
        """
        xyz, r = self._process_position(pos)

        R = self.radial(r, self.bas_n, self.bas_exp)

        Y = self.harmonics(xyz)
        return self._ao_kernel(R, Y)

    def _ao_kernel(self, R, Y):
        """Kernel for the ao values

        Args:
            R (torch.tensor): radial part of the AOs
            Y (torch.tensor): harmonics part of the AOs

        Returns:
            torch.tensor: values of the AOs (with contraction)
        """
        ao = self.norm_cst * R * Y
        if self.contract:
            ao = self._contract(ao)
        return ao

    def _compute_first_derivative_ao_values(self, pos, sum_grad):
        """Compute the value of the derivative of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, Nelec x Ndim
            sum_grad (boolean): return the sum_grad (True) or gradient (False)

        Returns:
            torch.tensor: derivative of atomic orbital values
                          size (Nbatch, Nelec, Norb) if sum_grad
                          size (Nbatch, Nelec, Norb, Ndim) if sum_grad=False
        """
        if sum_grad:
            return self._compute_sum_gradient_ao_values(pos)
        else:
            return self._compute_gradient_ao_values(pos)

    def _compute_sum_gradient_ao_values(self, pos):
        """Compute the jacobian of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, Nelec x Ndim

        Returns:
            torch.tensor: derivative of atomic orbital values
                          size (Nbatch, Nelec, Norb)

        """

        xyz, r = self._process_position(pos)

        R, dR = self.radial(r, self.bas_n,
                            self.bas_exp, xyz=xyz,
                            derivative=[0, 1])

        Y, dY = self.harmonics(xyz, derivative=[0, 1])

        return self._sum_gradient_kernel(R, dR, Y, dY)

    def _sum_gradient_kernel(self, R, dR, Y, dY):
        """Kernel for the jacobian of the ao values

        Args:
            R (torch.tensor): radial part of the AOs
            dR (torch.tensor): derivative of the radial part of the AOs
            Y (torch.tensor): harmonics part of the AOs
            dY (torch.tensor): derivative of the harmonics part of the AOs

        Returns:
            torch.tensor: values of the jacobian of the AOs (with contraction)
        """
        dao = self.norm_cst * (dR * Y + R * dY)
        if self.contract:
            dao = self._contract(dao)
        return dao

    def _compute_gradient_ao_values(self, pos):
        """Compute the gradient of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, Nelec x Ndim

        Returns:
            torch.tensor: derivative of atomic orbital values
                          size (Nbatch, Nelec, Norb, Ndim)

        """
        xyz, r = self._process_position(pos)

        R, dR = self.radial(r, self.bas_n,
                            self.bas_exp, xyz=xyz,
                            derivative=[0, 1],
                            sum_grad=False)

        Y, dY = self.harmonics(xyz, derivative=[0, 1], sum_grad=False)

        return self._gradient_kernel(R, dR, Y, dY)

    def _gradient_kernel(self, R, dR, Y, dY):
        """Kernel for the gradient of the ao values

        Args:
            R (torch.tensor): radial part of the AOs
            dR (torch.tensor): derivative of the radial part of the AOs
            Y (torch.tensor): harmonics part of the AOs
            dY (torch.tensor): derivative of the harmonics part of the AOs

        Returns:
            torch.tensor: values of the gradient of the AOs (with contraction)
        """
        nbatch = R.shape[0]
        bas = dR * Y.unsqueeze(-1) + R.unsqueeze(-1) * dY

        bas = self.norm_cst.unsqueeze(-1) * \
            self.bas_coeffs.unsqueeze(-1) * bas

        if self.contract:
            ao = torch.zeros(nbatch, self.nelec, self.norb,
                             3, device=self.device).type(torch.get_default_dtype())
            ao.index_add_(2, self.index_ctr, bas)
        else:
            ao = bas
        return ao

    def _compute_second_derivative_ao_values(self, pos, sum_hess):
        """Compute the values of the 2nd derivative of the ao from the xyz and r tensors

        Args:
            pos (torch.tensor): position of each elec size Nbatch, Nelec x Ndim
            sum_hess (boolean): return the sum_hess (True) or gradient (False)

        Returns:
            torch.tensor: derivative of atomic orbital values
                          size (Nbatch, Nelec, Norb) if sum_hess
                          size (Nbatch, Nelec, Norb, Ndim) if sum_hess=False
        """
        if sum_hess:
            return self._compute_sum_diag_hessian_ao_values(pos)
        else:
            return self._compute_diag_hessian_ao_values(pos)

    def _compute_sum_diag_hessian_ao_values(self, pos):
        """Compute the laplacian of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, Nelec x Ndim

        Returns:
            torch.tensor: derivative of atomic orbital values
                          size (Nbatch, Nelec, Norb)

        """
        xyz, r = self._process_position(pos)

        R, dR, d2R = self.radial(r, self.bas_n, self.bas_exp,
                                 xyz=xyz, derivative=[0, 1, 2],
                                 sum_grad=False)

        Y, dY, d2Y = self.harmonics(xyz,
                                    derivative=[0, 1, 2],
                                    sum_grad=False)
        return self._sum_diag_hessian_kernel(R, dR, d2R, Y, dY, d2Y)

    def _sum_diag_hessian_kernel(self, R, dR, d2R, Y, dY, d2Y):
        """Kernel for the sum of the diag hessian of the ao values

        Args:
            R (torch.tensor): radial part of the AOs
            dR (torch.tensor): derivative of the radial part of the AOs
            d2R (torch.tensor): 2nd derivative of the radial part of the AOs
            Y (torch.tensor): harmonics part of the AOs
            dY (torch.tensor): derivative of the harmonics part of the AOs
            d2Y (torch.tensor): 2nd derivative of the harmonics part of the AOs

        Returns:
            torch.tensor: values of the laplacian of the AOs (with contraction)
        """

        d2ao = self.norm_cst * \
            (d2R * Y + 2. * (dR * dY).sum(3) + R * d2Y)
        if self.contract:
            d2ao = self._contract(d2ao)
        return d2ao

    def _compute_diag_hessian_ao_values(self, pos):
        """Compute the individual elements of the laplacian of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, Nelec x Ndim

        Returns:
            torch.tensor: derivative of atomic orbital values
                          size (Nbatch, Nelec, Norb, 3)

        """

        xyz, r = self._process_position(pos)

        R, dR, d2R = self.radial(r, self.bas_n, self.bas_exp,
                                 xyz=xyz, derivative=[0, 1, 2],
                                 sum_grad=False, sum_hess=False)

        Y, dY, d2Y = self.harmonics(xyz,
                                    derivative=[0, 1, 2],
                                    sum_grad=False, sum_hess=False)

        return self._diag_hessian_kernel(R, dR, d2R, Y, dY, d2Y)

    def _diag_hessian_kernel(self, R, dR, d2R, Y, dY, d2Y):
        """Kernel for the diagonal hessian of the ao values

        Args:
            R (torch.tensor): radial part of the AOs
            dR (torch.tensor): derivative of the radial part of the AOs
            d2R (torch.tensor): 2nd derivative of the radial part of the AOs
            Y (torch.tensor): harmonics part of the AOs
            dY (torch.tensor): derivative of the harmonics part of the AOs
            d2Y (torch.tensor): 2nd derivative of the harmonics part of the AOs

        Returns:
            torch.tensor: values of the laplacian of the AOs (with contraction)
        """

        nbatch = R.shape[0]

        bas = self.norm_cst.unsqueeze(-1) * self.bas_coeffs.unsqueeze(-1) * \
            (d2R * Y.unsqueeze(-1) + 2. *
             (dR * dY) + R.unsqueeze(-1) * d2Y)

        if self.contract:
            d2ao = torch.zeros(nbatch, self.nelec, self.norb,
                               3, device=self.device).type(torch.get_default_dtype())
            d2ao.index_add_(2, self.index_ctr, bas)
        else:
            d2ao = bas

        return d2ao

    def _compute_mixed_second_derivative_ao_values(self, pos):
        """Compute the mixed second derivative of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, Nelec x Ndim

        Returns:
            torch.tensor: derivative of atomic orbital values
                          size (Nbatch, Nelec, Norb)

        """
        xyz, r = self._process_position(pos)

        R, dR, d2R, d2mR = self.radial(r, self.bas_n, self.bas_exp,
                                       xyz=xyz, derivative=[
                                           0, 1, 2, 3],
                                       sum_grad=False)

        Y, dY, d2Y, d2mY = self.harmonics(xyz,
                                          derivative=[0, 1, 2, 3],
                                          sum_grad=False)

        return self._off_diag_hessian_kernel(R, dR, d2R, d2mR, Y, dY, d2Y, d2mY)

    def _off_diag_hessian_kernel(self, R, dR, d2R, d2mR, Y, dY, d2Y, d2mY):
        """Kernel for the off diagonal hessian of the ao values

        Args:
            R (torch.tensor): radial part of the AOs
            dR (torch.tensor): derivative of the radial part of the AOs
            d2R (torch.tensor): 2nd derivative of the radial part of the AOs
            d2mR (torch.tensor): mixed 2nd derivative of the radial part of the AOs
            Y (torch.tensor): harmonics part of the AOs
            dY (torch.tensor): derivative of the harmonics part of the AOs
            d2Y (torch.tensor): 2nd derivative of the harmonics part of the AOs
            d2mY (torch.tensor): 2nd mixed derivative of the harmonics part of the AOs

        Returns:
            torch.tensor: values of the mixed derivative of the AOs (with contraction)
        """

        nbatch = R.shape[0]

        bas = self.norm_cst.unsqueeze(-1) * self.bas_coeffs.unsqueeze(-1) * \
            (d2mR * Y.unsqueeze(-1) +
             ((dR[..., [[0, 1], [0, 2], [1, 2]]] *
               dY[..., [[1, 0], [2, 0], [2, 1]]]).sum(-1))
             + R.unsqueeze(-1) * d2mY)

        if self.contract:
            d2ao = torch.zeros(nbatch, self.nelec, self.norb,
                               3, device=self.device).type(torch.get_default_dtype())
            d2ao.index_add_(2, self.index_ctr, bas)
        else:
            d2ao = bas

        return d2ao

    def _compute_all_ao_values(self, pos):
        """Compute the ao, gradient, laplacian of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, Nelec x Ndim
            sum_grad (bool): return the sum of the gradients if True
            sum_hess (bool): returns the sum of the diag hess if True
        Returns:
            tuple(): (ao, grad and lapalcian) of atomic orbital values
                     ao size (Nbatch, Nelec, Norb)
                     dao size (Nbatch, Nelec, Norb, Ndim)
                     d2ao size (Nbatch, Nelec, Norb)

        """

        xyz, r = self._process_position(pos)

        # the gradients elements are needed to compute the second der
        # we therefore use sum_grad=False regardless of the input arg
        R, dR, d2R = self.radial(r, self.bas_n, self.bas_exp,
                                 xyz=xyz, derivative=[0, 1, 2],
                                 sum_grad=False)

        # the gradients elements are needed to compute the second der
        # we therefore use sum_grad=False regardless of the input arg
        Y, dY, d2Y = self.harmonics(xyz,
                                    derivative=[0, 1, 2],
                                    sum_grad=False)

        return (self._ao_kernel(R, Y),
                self._gradient_kernel(R, dR, Y, dY),
                self._sum_diag_hessian_kernel(R, dR, d2R, Y, dY, d2Y))

    def _process_position(self, pos):
        """Computes the positions/distance bewteen elec/orb

        Args:
            pos (torch.tensor): positions of the walkers Nbat, NelecxNdim

        Returns:
            torch.tensor, torch.tensor: positions of the elec wrt the bas
                                        (Nbatch, Nelec, Norb, Ndim)
                                        distance between elec and bas
                                        (Nbatch, Nelec, Norb)
        """
        # get the elec-atom vectors/distances
        xyz, r = self._elec_atom_dist(pos)

        # repeat/interleave to get vector and distance between
        # electrons and orbitals
        return (xyz.repeat_interleave(self.nshells, dim=2),
                r.repeat_interleave(self.nshells, dim=2))

    def _elec_atom_dist(self, pos):
        """Computes the positions/distance bewteen elec/atoms

        Args:
            pos (torch.tensor): positions of the walkers : Nbatch x [Nelec*Ndim]

        Returns:
            (torch.tensor, torch.tensor): positions of the elec wrt the atoms
                                        [Nbatch x Nelec x Natom x Ndim]
                                        distance between elec and atoms
                                        [Nbatch x Nelec x Natom]
        """

        # compute the vectors between electrons and atoms
        xyz = (pos.view(-1, self.nelec, 1, self.ndim) -
               self.atom_coords[None, ...])

        # distance between electrons and atoms
        r = torch.sqrt((xyz*xyz).sum(3))

        return xyz, r

    def _contract(self, bas):
        """Contrat the basis set to form the atomic orbitals

        Args:
            bas (torch.tensor): values of the basis function

        Returns:
            torch.tensor: values of the contraction
        """
        nbatch = bas.shape[0]
        bas = self.bas_coeffs * bas
        cbas = torch.zeros(nbatch, self.nelec,
                           self.norb, device=self.device
                           ).type(torch.get_default_dtype())
        cbas.index_add_(2, self.index_ctr, bas)
        return cbas

    def update(self, ao, pos, idelec):
        """Update an AO matrix with the new positions of one electron

        Args:
            ao (torch.tensor): initial AO matrix
            pos (torch.tensor): new positions of some electrons
            idelec (int): index of the electron that has moved

        Returns:
            torch.tensor: new AO matrix

        Examples::
            >>> mol = Molecule('h2.xyz')
            >>> ao = AtomicOrbitals(mol)
            >>> pos = torch.rand(100,6)
            >>> aovals = ao(pos)
            >>> id = 0
            >>> pos[:,:3] = torch.rand(100,3)
            >>> ao.update(aovals, pos, 0)
        """

        ao_new = ao.clone()
        ids, ide = (idelec) * 3, (idelec + 1) * 3
        ao_new[:, idelec, :] = self.forward(
            pos[:, ids:ide], one_elec=True).squeeze(1)
        return ao_new



import numpy as np
from scipy.optimize import curve_fit


def blocking(x, block_size, expand=False):
    """block the data

    Args:
        x (data): size Nsample, Nexp
        block_size (int): size of the block
    """
    nstep, nwalkers = x.shape
    nblock = nstep // block_size

    xb = np.copy(x[:block_size * nblock, :])
    xb = xb.reshape(nblock, block_size, nwalkers).mean(axis=1)

    if expand:
        xb = xb.T.repeat(block_size).reshape(nwalkers, -1).T[:nstep]

    return xb


def correlation_coefficient(x, norm=True):
    """Computes the correlation coefficient

    Args:
        x (np.ndarray): measurement of size [Nsample, Nexperiments]
        norm (bool, optional): [description]. Defaults to True.
    """

    N = x.shape[0]
    xm = x-x.mean(0)

    c = np.zeros_like(x)
    for tau in range(0, N):
        c[tau] = 1./(N-tau) * (xm[:N-tau] * xm[tau:]).sum(0)

    if norm:
        c /= c[0]

    return c


def integrated_autocorrelation_time(correlation_coeff, size_max):
    """Computes the integrated autocorrelation time

    Args:
        correlation_coeff (np.ndarray): coeff size Nsample,Nexp
        size_max (int): max size 
    """
    return 1. + 2. * np.cumsum(correlation_coeff[1:size_max], 0)


def fit_correlation_coefficient(coeff):
    """Fit the correlation coefficient 
       to get the correlation time.

    Args:
        coeff (np.ndarray): correlation coefficient

    Returns:
        float, np.ndarray: correlation time, fitted curve
    """

    def fit_exp(x, y):
        """Fit an exponential to the data."""
        def func(x, tau):
            return np.exp(-x/tau)
        popt, pcov = curve_fit(func, x, y, p0=(1.))
        return popt[0], func(x, popt)

    return fit_exp(np.arange(len(coeff)), coeff)



import unittest

import numpy as np
import torch
import torch.optim as optim

from qmctorch.sampler import Metropolis
from qmctorch.solver import Solver
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow

from ..path_utils import PATH_TEST


class TestH2ADF(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(0)

        # molecule
        path_hdf5 = (
            PATH_TEST / 'hdf5/H2_adf_dzp.hdf5').absolute().as_posix()
        self.mol = Molecule(load=path_hdf5)

        # wave function
        self.wf = SlaterJastrow(self.mol, kinetic='auto',
                                configs='single(2,2)')

        # sampler
        self.sampler = Metropolis(
            nwalkers=1000,
            nstep=2000,
            step_size=0.5,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        # optimizer
        self.opt = optim.Adam(self.wf.parameters(), lr=0.01)

        # solver
        self.solver = Solver(wf=self.wf, sampler=self.sampler,
                                          optimizer=self.opt)

        # ground state energy
        self.ground_state_energy = -1.16

        # ground state pos
        self.ground_state_pos = 0.69

    def test_single_point(self):

        self.solver.wf.ao.atom_coords[0, 2] = -self.ground_state_pos
        self.solver.wf.ao.atom_coords[1, 2] = self.ground_state_pos
        self.solver.sampler = self.sampler

        # sample and compute observables
        obs = self.solver.single_point()
        e, v = obs.energy, obs.variance

        # vals on different archs
        expected_energy = [-1.1572532653808594,
                           -1.1501641653648578]

        expected_variance = [0.05085879936814308,
                             0.05094174843043177]

        assert(np.any(np.isclose(e.data.item(), np.array(expected_energy))))
        assert(np.any(np.isclose(v.data.item(), np.array(expected_variance))))

        # assert(e > 2 * self.ground_state_energy and e < 0.)
        # assert(v > 0 and v < 5.)

    def test_wf_opt_auto_grad(self):

        self.solver.configure(track=['local_energy'],
                              loss='energy', grad='auto')
        obs = self.solver.run(5)

    def test_wf_opt_manual_grad(self):
        self.solver.configure(track=['local_energy'],
                              loss='energy', grad='manual')
        obs = self.solver.run(5)


if __name__ == "__main__":
    unittest.main()
    # t = TestH2ADF()
    # t.setUp()
    # t.test_single_point()
